<?php
/**
 * @noinspection PhpMissingReturnTypeInspection
 */

namespace SV\AdvancedBbCodesPack\BbCode;

use XF\BbCode\Parser;
use XF\BbCode\RuleSet;

/**
 * Class ChildTagParser
 *
 * @package SV\AdvancedBbCodesPack\BbCode
 */
class ChildTagParser extends Parser
{
    /**
     * @var string
     */
    protected $openBracket = '{';

    /**
     * @var string
     */
    protected $closeBracket = '}';

    /**
     * @param string|\XF\Phrase $text
     * @param RuleSet $ruleSet
     * @return array
     * @noinspection RegExpUnnecessaryNonCapturingGroup
     * @noinspection RegExpRedundantEscape
     * @noinspection PregQuoteUsageInspection
     */
    public function parse($text, RuleSet $ruleSet)
    {
        $this->ruleSet = $ruleSet;

        $this->ast = [];
        $this->astReference = &$this->ast;
        $this->tagStack = [];
        $this->pendingText = '';
        $this->plainTag = null;
        $this->depth = 0;

        $position = 0;
        $text = (string)$text;
        $length = \strlen($text);

        $openBracket = $this->openBracket;
        $closeBracket = $this->closeBracket;

        $openBracketRegex = \preg_quote($openBracket, '#');
        $closeBracketRegex = \preg_quote($closeBracket, '#');

        $regex = '#(?:' . $openBracketRegex . '([a-z\d_]+)(=|' . $closeBracketRegex . ')|' . $openBracketRegex . '/([a-z\d_]+)' . $closeBracketRegex . ')#i';

        while (\preg_match($regex, $text, $match, PREG_OFFSET_CAPTURE, $position))
        {
            if ($match[0][1] > $position)
            {
                // push text
                $plainText = \substr($text, $position, $match[0][1] - $position);
                $this->pushText($plainText);
            }

            $fullMatch = $match[0][0];
            $position = $match[0][1] + \strlen($fullMatch);

            if (isset($match[3]))
            {
                $this->closeTag($fullMatch, $match[3][0]);
            }
            else if ($match[2][0] === $closeBracket)
            {
                // simple, optionless tag
                $this->openTag($fullMatch, $match[1][0]);
            }
            else
            {
                // complex tag
                if ($position >= $length)
                {
                    $this->pushText($fullMatch);
                }
                else
                {
                    $delim = \substr($text, $position, 1);
                    if ($delim === '"' || $delim === "'")
                    {
                        $startPos = $position + 1;
                        $endPos = \strpos($text, "$delim}", $startPos);
                        $startMatch = $delim;
                        $endMatch = $delim . $closeBracket;
                    }
                    else
                    {
                        $startPos = $position;
                        $endPos = \strpos($text, $closeBracket, $startPos);
                        $startMatch = '';
                        $endMatch = $closeBracket;
                    }
                    if ($endPos)
                    {
                        $option = \substr($text, $startPos, $endPos - $startPos);
                        $this->openTag(
                            $fullMatch . $startMatch . $option . $endMatch,
                            $match[1][0],
                            $option
                        );

                        $position = $endPos + \strlen($endMatch);
                    }
                    else
                    {
                        $this->pushText($fullMatch);
                    }
                }
            }
        }

        if ($position < $length)
        {
            $this->pushText(\substr($text, $position));
        }

        $this->finalizeText();

        $ast = $this->ast;
        $this->ast = [];

        $null = null;
        $this->astReference = &$null;

        return $ast;
    }

    /**
     * @param string $originalText
     * @param string $tag
     * @param string $option
     */
    protected function openTag($originalText, $tag, $option = null)
    {
        if ($this->plainTag)
        {
            $this->pushText($originalText);

            return;
        }

        $tagLower = \strtolower($tag);
        if (
            !$this->ruleSet->validateTag($tagLower, $option, $modifiers)
            || !$this->hasMatchingParent($tagLower)
        )
        {
            $this->pushText($originalText);

            return;
        }

        $this->finalizeText();

        $i = \count($this->astReference);
        $this->astReference[$i] = [
            'tag'      => $tagLower,
            'option'   => $option,
            'original' => [
                $originalText,
                "{$this->openBracket}/{$tag}{$this->closeBracket}"
            ],
            'children' => []
        ];

        \array_unshift($this->tagStack, [
            'tag'       => $tagLower,
            'parent'    => &$this->astReference,
            'entry'     => &$this->astReference[$i],
            'modifiers' => $modifiers
        ]);

        $this->astReference = &$this->astReference[$i]['children'];

        if ($modifiers['plain'] ?? false)
        {
            $this->plainTag = $tagLower;
        }

        $this->depth++;
    }

    /**
     * @param string $tagLower
     * @return bool
     */
    protected function hasMatchingParent(string $tagLower): bool
    {
        $definition = $this->ruleSet->getTag($tagLower);
        if (!\is_array($definition))
        {
            return false;
        }

        if (isset($definition['allowedParents']))
        {
            $parent = $this->tagStack ?  $this->tagStack[0]['tag'] : '';

            if (!\in_array($parent, $definition['allowedParents'], true))
            {
                return false;
            }
        }

        return true;
    }
}