<?php

namespace SV\AdvancedBbCodesPack\BbCode\Renderer;

use DOMDocument;
use XF\BbCode\Renderer\AbstractRenderer;
use XF\Str\Formatter;
use XF\Template\Templater;

/**
 * Class TableHtml
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Renderer
 */
class TableHtml extends AbstractRenderer
{
    use StyleTrait;

    /**
     * @var array
     */
    protected $tagOptions;

    /**
     * @var Formatter
     */
    protected $formatter;

    /**
     * @var Templater
     */
    protected $templater;

    /**
     * @var \XF\App
     */
    protected $app;

    /**
     * @var bool
     */
    protected $disableTextNodesAfter = false;

    /**
     * @var array
     */
    protected $tableOptions;

    /**
     * @var array
     */
    protected $tableStyle;

    /**
     * @var \XF\BbCode\Renderer\AbstractRenderer|null
     */
    protected $parentRenderer;

    /** @var bool */
    protected $tidyHTML = false;

    /**
     * @param Formatter $formatter
     * @param Templater $templater
     */
    public function __construct(Formatter $formatter, Templater $templater)
    {
        $this->app = \XF::app();
        $this->formatter = $formatter;
        $this->templater = $templater;
        $this->addDefaultTags();
    }

    public function isTidyHTML(): bool
    {
        return $this->tidyHTML;
    }

    public function setTidyHTML(bool $tidyHTML)
    {
        $this->tidyHTML = $tidyHTML;
    }

    public function setTemplater(Templater $templater = null)
    {
        $this->templater = $templater;
    }

    /**
     * @param AbstractRenderer|null $parentRenderer
     */
    public function setParentRender(\XF\BbCode\Renderer\AbstractRenderer $parentRenderer = null)
    {
        $this->parentRenderer = $parentRenderer;
    }

    /**
     * @return AbstractRenderer|null
     */
    public function getParentRender()
    {
        return $this->parentRenderer;
    }

    /**
     * @param \XF\App $app
     * @return static
     */
    public static function factory(\XF\App $app): TableHtml
    {
        return new static($app->stringFormatter(), $app->templater());
    }

    public function getCustomTagConfig(array $tag): array
    {
        return [];
    }

    public function addDefaultTags()
    {
        $style = $this->getStyle();

        $applyWidthRestrictions = $style->getProperty('svBbcodeTableApplyWidthRestrictions');
        $applyHeightRestrictions = $style->getProperty('svBbcodeTableApplyHeightRestrictions ');

        $this->tableStyle = [
            'size'   => [
                'px'      => [
                    'minWidth'  => $applyWidthRestrictions ? $style->getProperty('svBbcodeTableMinWidthPx') : null,
                    'maxWidth'  => $applyWidthRestrictions ? $style->getProperty('svBbcodeTableMaxWidthPx') : null,
                    'minHeight' => $applyHeightRestrictions ? $style->getProperty('svBbcodeTableMinHeightPx') : null,
                    'maxHeight' => $applyHeightRestrictions ? $style->getProperty('svBbcodeTableMaxHeightPx') : null,
                ],
                'percent' => [
                    'minWidth'  => $applyWidthRestrictions ? $style->getProperty('svBbcodeTableMinWidthPercent') : null,
                    'maxWidth'  => $applyWidthRestrictions ? $style->getProperty('svBbcodeTableMaxWidthPercent') : null,
                    'minHeight' => $applyHeightRestrictions ? $style->getProperty('svBbcodeTableMinHeightPercent') : null,
                    'maxHeight' => $applyHeightRestrictions ? $style->getProperty('svBbcodeTableMaxHeightPercent') : null,
                ]
            ],
            'cell'   => [
                'maxCellpadding' => $style->getProperty('svBbcodeTableMaxCellpadding'),
                'maxCellspacing' => $style->getProperty('svBbcodeTableMaxCellspacing')
            ],
            'border' => [
                'max' => $style->getProperty('svBbcodeTableMaxBorderWidth')
            ]
        ];

        $this->tagOptions = [
            'xtable'   => [
                'tableTag' => true,
                'disableTextNodesIn' => true,
                'options' => [
                    'size',
                    'fright',
                    'fleft',
                    'bright',
                    'bcenter',
                    'bleft',
                    'bcolor',
                    'cellspacing',
                    'cellpadding',
                    'border',
                    'border-color',
                ],
            ],
            'thead'    => [
                'disableTextNodesIn' => true,
                'options'            => [
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline'
                ],
            ],
            'tbody'    => [
                'disableTextNodesIn' => true,
                'options'            => [
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline'
                ],
            ],
            'tfoot'    => [
                'disableTextNodesIn' => true,
                'options'            => [
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline'
                ],
            ],
            'colgroup' => [
                'disableTextNodesIn' => true,
                'options'            => [
                    'size',
                    'bcolor',
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline'
                ],
            ],
            'col'      => [
                'options' => [
                    'size',
                    'span',
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline'
                ],
            ],
            'caption'  => [
                'options' => [
                    'size',
                    'bcolor',
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline'
                ],
            ],
            'tr'       => [
                'disableTextNodesIn' => true,
                'options'            => [
                    'size',
                    'bcolor',
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline',
                    'border',
                    'border-color',
                ],
            ],
            'td'       => [
                'options' => [
                    'size',
                    'colspan',
                    'rowspan',
                    'bcolor',
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline',
                    'border',
                    'border-color',
                    'scope',
                    'nowrap',
                    'bright',
                    'bcenter',
                    'bleft',
                ],
            ],
            'th'       => [
                'options' => [
                    'size',
                    'colspan',
                    'rowspan',
                    'bcolor',
                    'left',
                    'center',
                    'right',
                    'bottom',
                    'middle',
                    'top',
                    'baseline',
                    'border',
                    'border-color',
                    'scope',
                    'nowrap',
                ],
            ]
        ];

        $this->tags = $this->tagOptions;
    }

    public function addTag($tag, array $config)
    {
        // block adding custom tags in the child parser
    }

    /**
     * @param string $tagName
     * @param string $key
     * @param string $value
     */
    protected function unknownTagOptionKey(string $tagName, string $key, string $value)
    {
        if ((\XF::options()->svAdvancedBbCodeLogUnknownOption ?? false) || \XF::$developmentMode)
        {
            $s = "Tag {$tagName} - option; $key:{$value} is no implemented/unknown";
            if (\XF::$developmentMode)
            {
                throw new \BadMethodCallException($s);
            }
            \XF::logError($s);
        }
    }

    /**
     * @param bool     $isMaster
     * @param string   $tagName
     * @param string[] $tagOptions
     * @return string[]
     */
    public function parseTagOption(bool $isMaster, string $tagName, array $tagOptions): array
    {
        if (!$tagOptions || empty($this->tagOptions[$tagName]['options']))
        {
            return ['', '', ''];
        }

        $extraClass = [];
        $attributes = [];
        $css = [];
        $options = $this->tagOptions[$tagName]['options'];
        $optionsStack = [];
        foreach ($tagOptions as $key => $value)
        {
            if (\is_integer($key))
            {
                $lValue = \strtolower($value);
                if (\in_array($lValue, $options, true))
                {
                    $key = $lValue;
                }
                // this may be a size-like thing
                else if (empty($optionsStack['size']) && \in_array('size', $options, true))
                {
                    if ($this->parseSizeArgs($isMaster, $tagName, 'size', $value, $optionsStack, $css, false))
                    {
                        continue;
                    }
                }
            }
            else
            {
                $key = \strtolower($key);
            }

            if (isset($optionsStack[$key]))
            {
                continue;
            }

            if (!\in_array($key, $options, true))
            {
                if ($isMaster &&
                    \preg_match('/^skin\d{1,2}$/i', $value, $match))
                {
                    if (empty($optionsStack['theme_class']))
                    {
                        $class = $match[0];
                        $extraClass[] = $class;
                        $optionsStack['theme_class'] = true;
                        continue;
                    }
                }
                else
                {
                    $this->unknownTagOptionKey($tagName, $key, $value);
                }
            }

            $setOptionStack = false;
            switch ($key)
            {
                case 'size':
                    $this->parseSizeArgs($isMaster, $tagName, $key, $value, $optionsStack,$css, true);
                    break;

                case 'fright':
                    $css[] = 'float: right';
                    $setOptionStack = true;
                    break;

                case 'fleft':
                    $css[] = 'float: left';
                    $setOptionStack = true;
                    break;

                case 'bright':
                    $css[] = 'margin-left: auto';
                    $css[] = 'margin-right: 0';
                    $setOptionStack = true;
                    break;

                case 'bcenter':
                    $css[] = 'margin-left: auto';
                    $css[] = 'margin-right: auto';
                    $setOptionStack = true;
                    break;

                case 'bleft':
                    $css[] = 'margin-left: 0';
                    $css[] = 'margin-right: auto';
                    $setOptionStack = true;
                    break;

                case 'cellpadding':
                    $regex = '/^(\d{1,2})$/i';
                    if (\preg_match($regex, $value, $match))
                    {
                        $value = $match[1];
                        $params = $this->tableStyle['cell'];

                        $value = ($value > $params['maxCellpadding'])
                            ? $params['maxCellpadding']
                            : $value;
                        $unit = ($value === 0) ? '' : 'px';

                        $css[] = "padding: {$value}{$unit}";
                        $attributes[] = "cellpadding='{$value}'";
                        $setOptionStack = true;
                    }
                    break;

                case 'cellspacing':
                    $regex = '/^(\d{1,2})$/i';
                    if (\preg_match($regex, $value, $match))
                    {
                        $value = $match[1];
                        $params = $this->tableStyle['cell'];

                        $value = ($value > $params['maxCellspacing'])
                            ? $params['maxCellspacing']
                            : $value;
                        $unit = ($value === 0) ? '' : 'px';

                        $css[] = "border-spacing: {$value}{$unit}";
                        $attributes[] = "cellspacing='{$value}'";

                        $setOptionStack = true;
                    }
                    break;

                case 'bcolor':
                    $regex = '/^(rgb\(\s*\d+%?\s*,\s*\d+%?\s*,\s*\d+%?\s*\)|#[a-f\d]{6}|#[a-f\d]{3}|[a-z]+)$/i';

                    if (\preg_match($regex, $value, $match))
                    {
                        $bgColor = $match[1];

                        $css[] = "background-color: $bgColor";
                        $setOptionStack = true;
                    }
                    break;

                case 'border':
                    $regex = '/^(\d{1,2})$/i';
                    if (\preg_match($regex, $value, $match))
                    {
                        $value = $match[1];
                        $params = $this->tableStyle['border'];

                        $value = ($value > $params['max']) ? $params['max'] : $value;
                        $value = (int)$value;
                        $unit = ($value === 0) ? '' : 'px';

                        $css[] = "border-width: {$value}{$unit}";
                        $attributes[] = "border='{$value}'";
                        $setOptionStack = true;
                    }
                    break;

                case 'border-color':
                    $regex = '/^(rgb\(\s*\d+%?\s*,\s*\d+%?\s*,\s*\d+%?\s*\)|#[a-f\d]{6}|#[a-f\d]{3}|[a-z]+)$/i';

                    if (\preg_match($regex, $value, $match))
                    {
                        $bdColor = $match[1];

                        $css[] = "border-color: $bdColor";
                        $setOptionStack = true;
                    }
                    break;

                case 'left':
                    $css[] = 'text-align: left';
                    $setOptionStack = true;
                    break;

                case 'center':
                    $css[] = 'text-align: center';
                    $setOptionStack = true;
                    break;

                case 'right':
                    $css[] = 'text-align: right';
                    $setOptionStack = true;
                    break;

                case 'bottom':
                    $css[] = 'vertical-align: bottom';
                    $setOptionStack = true;
                    break;

                case 'middle':
                    $css[] = 'vertical-align: middle';
                    $setOptionStack = true;
                    break;

                case 'top':
                    $css[] = 'vertical-align: top';
                    $setOptionStack = true;
                    break;

                case 'baseline':
                    $css[] = 'vertical-align: baseline';
                    $setOptionStack = true;
                    break;

                case 'span':
                    //http://www.w3schools.com/tags/att_col_span.asp
                    $regex = '/^(\d{1,2})$/i';

                    if (\preg_match($regex, $value, $match))
                    {
                        $value = $match[1];
                        /*Checker not really needed here, the regex is enough*/

                        $attributes[] = "span='{$value}'";
                        $setOptionStack = true;
                    }
                    break;

                case 'colspan':
                    $regex = '/^(\d{1,2})$/i';

                    if (\preg_match($regex, $value, $match))
                    {
                        $value = $match[1];
                        /*Checker not really needed here, the regex is enough*/

                        $attributes[] = "colspan='{$value}'";
                        $setOptionStack = true;
                    }
                    break;

                case 'rowspan':
                    $regex = '/^(\d{1,2})$/i';

                    if (\preg_match($regex, $value, $match))
                    {
                        $value = $match[1];
                        /*Checker not really needed here, the regex is enough*/

                        $attributes[] = "rowspan='{$value}'";
                        $setOptionStack = true;
                    }
                    break;

                case 'scope':
                    $regex = '/^(\d{1,2})$/i';

                    if (\preg_match($regex, $value, $match))
                    {
                        $value = $match[1];
                        /*Checker not really needed here, the regex is enough*/

                        $attributes[] = "scope='{$value}'";
                        $setOptionStack = true;
                    }
                    break;

                case 'nowrap':
                    $css[] = 'white-space: nowrap';
                    $attributes[] = "nowrap='{$value}'";
                    $setOptionStack = true;
                    break;

                default:
                    $this->unknownTagOptionKey($tagName, $key, $value);
                    break;
            }

            if ($setOptionStack)
            {
                $optionsStack[$key] = true;
            }
        }

        $attributes = $attributes ? ' ' . \implode(' ', $attributes) : '';
        $css = \implode('; ', $css) ?: '';
        $extraClass = $extraClass ? ' ' . \implode(' ', $extraClass) : '';

        return [$attributes, $css, $extraClass];
    }


    /**
     * @param bool   $isMaster
     * @param string $tagName
     * @param string $key
     * @param string $value
     * @param array  $optionsStack
     * @param array  $css
     * @param bool   $reportUnknown
     * @return bool
     */
    protected function parseSizeArgs(bool $isMaster, string $tagName, string $key, string $value, array &$optionsStack, array &$css, bool $reportUnknown): bool
    {
        $regex = '#^(\d+(?:\.\d+)?|@)(|%|px)x(\d+(?:\.\d+)?|@)(|%|px)$#';

        if (\preg_match($regex, $value, $match))
        {
            $width = $match[1];
            $widthUnit = empty($match[2]) ? 'px' : $match[2];
            $height = $match[3];

            $heightUnit = empty($match[4]) ? 'px' : $match[4];

            if ($isMaster)
            {
                list($params_px, $params_percent) = \array_values($this->tableStyle['size']);

                if ($width !== '@')
                {
                    if ($widthUnit === 'px')
                    {
                        $maxWidth = $params_px['maxWidth'];
                        $minWidth = $params_px['minWidth'];
                    }
                    else
                    {
                        $maxWidth = $params_percent['maxWidth'];
                        $minWidth = $params_percent['minWidth'];
                    }

                    if ($maxWidth && $width > $maxWidth)
                    {
                        $width = $maxWidth;
                    }
                    else if ($minWidth && $width < $minWidth)
                    {
                        $width = $minWidth;
                    }
                }

                if ($height !== '@')
                {
                    if ($heightUnit === 'px')
                    {
                        $maxHeight = $params_px['maxHeight'];
                        $minHeight = $params_px['minHeight'];
                    }
                    else
                    {
                        $maxHeight = $params_percent['maxHeight'];
                        $minHeight = $params_percent['minHeight'];
                    }

                    if ($maxHeight && $height > $maxHeight)
                    {
                        $height = $maxHeight;
                    }
                    else if ($minHeight && $height < $minHeight)
                    {
                        $height = $minHeight;
                    }
                }
            }

            if ($width && $width !== '@')
            {
                $css[] = "width: {$width}{$widthUnit}";
            }

            if ($height && $height !== '@')
            {
                $css[] = "height: {$height}{$heightUnit}";
            }

            $optionsStack[$key] = true;

            return true;
        }
        else if ($reportUnknown)
        {
            $this->unknownTagOptionKey($tagName, $key, $value);
        }

        return false;
    }

    /**
     * @param array $tag
     * @param array $options
     * @return string|\XF\PreEscaped
     * @throws \Exception
     */
    public function renderTag(array $tag, array $options)
    {
        $options['disableTextOutput'] = false;

        $tagName = $tag['tag'];
        if (isset($this->tags[$tagName]))
        {
            if (empty($this->tagOptions[$tagName]['tableTag']))
            {
                return $this->renderChildTag($tag, $options);
            }

            return $this->renderTable($tag, $options);
        }

        return $this->renderUnparsedTag($tag, $options);
    }

    /**
     * @param array $tag
     * @param array $options
     * @return string|\XF\PreEscaped
     * @throws \Exception
     */
    public function renderTable(array $tag, array $options)
    {
        $tagName = $tag['tag'];
        $renderRule = $this->tags[$tagName];

        $tagOptions = $this->getTagOptions($tag['option'], $options);
        list($attributes, $css, $extraClass) = $this->parseTagOption(true, $tag['tag'], $tagOptions);

        $options['disableTextOutput'] = $renderRule['disableTextNodesIn'] ?? false;
        $content = $this->templater->preEscaped($this->renderSubTree($tag['children'], $options));

        return $this->templater->renderTemplate('public:sv_bbcode_table', [
            'rawTableHtml' => $content,
            'css'          => $css,
            'attributes'   => $attributes,
            'extraClass'   => $extraClass,
        ]);
    }

    /**
     * @param array            $children
     * @param string|array     $option
     * @param array            $tag
     * @param array            $options
     * @param AbstractRenderer $renderer
     * @return string|\XF\PreEscaped
     * @noinspection PhpUnusedParameterInspection
     */
    public function renderTagSvTableHtml(array $children, $option, array $tag, array $options, AbstractRenderer $renderer)
    {
        $output = '';

        foreach ($children as $element)
        {
            if (\is_array($element))
            {
                $output .= $renderer->renderTag($element, $options);
            }
            else
            {
                $output .= $element;
            }
        }

        return $output;
    }

    /**
     * @param array $tree
     * @param array $options
     * @return string|\XF\PreEscaped
     * @throws \Exception
     */
    public function renderSubTree(array $tree, array $options)
    {
        $enableTextOutput = empty($options['disableTextOutput']);

        $output = '';
        if ($enableTextOutput)
        {
            $wasNotSet = empty($options['deferrBbCodeParsing']);
            $hasPotentialWrappedBbCode = !$wasNotSet;
            if ($wasNotSet)
            {
                foreach ($tree as $key => $element)
                {
                    if (\is_string($element) &&
                        $key + 2 < \count($tree) &&
                        \is_array($tree[$key + 1]) &&
                        \is_string($tree[$key + 2]) &&
                        \strpos($element, '[') !== false)
                    {
                        $hasPotentialWrappedBbCode = true;
                        break;
                    }
                }
            }
            if ($hasPotentialWrappedBbCode)
            {
                $deferParsing = $options + [
                    'deferrBbCodeParsing' => true,
                ];

                foreach ($tree as $element)
                {
                    if (\is_array($element))
                    {
                        $output .= '[svtablehtml]' . $this->renderTag($element, $options) . '[/svtablehtml]';
                    }
                    else
                    {
                        $output .= $this->renderString($element, $deferParsing);
                    }
                }

                if ($wasNotSet)
                {
                    $parser = $this->app->bbCode()->parser();
                    $rules = $this->parentRenderer->getRules();
                    $rules->addTag('svtablehtml', []);
                    $this->parentRenderer->addTag('svtablehtml', [
                        'callback' => [$this, 'renderTagSvTableHtml'],
                    ]);
                    try
                    {
                        $ast = $parser->parse($output, $rules);
                        $output = $this->parentRenderer->renderSubTree($ast, $options);
                    }
                    finally
                    {
                        $rules->removeTag('svtablehtml');
                        $this->parentRenderer->removeTag('svtablehtml');
                    }
                }
            }
            else
            {
                foreach ($tree as $element)
                {
                    if (\is_array($element))
                    {
                        $output .= $this->renderTag($element, $options);
                    }
                    else
                    {
                        $output .= $this->renderString($element, $options);
                    }
                }
            }
        }
        else
        {
            foreach ($tree as $element)
            {
                if (\is_array($element))
                {
                    $output .= $this->renderTag($element, $options);
                }
            }
        }

        return $output;
    }

    /**
     * @param null|string|string[] $tagOption
     * @param array        $options
     * @return string[]
     */
    protected function getTagOptions($tagOption, array $options = []): array
    {
        if ($tagOption === null)
        {
            return [];
        }

        if (\is_array($tagOption))
        {
            return $tagOption;
        }

        $options['stopLineBreakConversion'] = 1 + (int)($options['stopLineBreakConversion'] ?? 0);
        $options['stopSmilies'] = 1 + (int)($options['stopSmilies'] ?? 0);

        $tagOption = $this->renderString($tagOption, $options);

        $tagOptionsRaw = \explode('|', $tagOption) ?: [];

        $tagOptions = [];
        foreach($tagOptionsRaw as $value)
        {
            if (\preg_match("#^\s*([a-zA-Z_-]+)\s*:\s*(.*)\s*$#us", $value, $matches))
            {
                if ($matches[2] === '' || \is_integer($matches[1]))
                {
                    $tagOptions[] = $matches[1];
                }
                else
                {
                    $tagOptions[$matches[1]] = $matches[2];
                }
            }
            else if ($value !== '')
            {
                $tagOptions[] = $value;
            }
        }

        return $tagOptions;
    }

    /**
     * @param array $tag
     * @param array $options
     * @return string|\XF\PreEscaped
     * @throws \Exception
     */
    protected function renderChildTag(array $tag, array $options)
    {
        $tagName = $tag['tag'];
        $renderRule = $this->tags[$tagName];

        $tagOptions = $this->getTagOptions($tag['option'], $options);
        list($attributes, $css, $extraClass) = $this->parseTagOption(false, $tagName, $tagOptions);

        $formattedClass = $extraClass ? " class='{$extraClass}'" : '';
        $formattedCss = $css ? " style='{$css}'" : '';

        $openingHtmlTag = "<{$tagName}{$formattedClass}{$attributes}{$formattedCss}>\n";
        $closingHtmlTag = "\n</$tagName>\n";

        $options['disableTextOutput'] = $renderRule['disableTextNodesIn'] ?? false;
        $content = $this->renderSubTree($tag['children'], $options);

        if ($content === '' || \ctype_space($content))
        {
            // Will avoid tags to be "eaten"
            $content = '&nbsp;';
        }

        return $openingHtmlTag . $content . $closingHtmlTag;
    }

    /**
     * @param string $string
     * @param array  $options
     * @return string|\XF\PreEscaped
     */
    public function renderString($string, array $options)
    {
        if (!$string || ($options['disableTextOutput'] ?? false))
        {
            return '';
        }

        if (!$this->parentRenderer)
        {
            if (\strpos($string, '<') === false)
            {
                // does not contain HTML like content
                return $string;
            }

            return $this->tidyHTML ? $this->tidyHTML($string) : $string;
        }

        if ($options['deferrBbCodeParsing'] ?? false)
        {
            return $string;
        }

        if (\strpos($string, '[') === false)
        {
            // does not contain bb-code like content
            return $this->parentRenderer->renderString($string, $options);
        }

        $parser = $this->app->bbCode()->parser();
        $ast = $parser->parse($string, $this->parentRenderer->getRules());

        return $this->parentRenderer->renderSubTree($ast, $options);
    }

    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    protected function tidyHTML(string $html): string
    {
        if (!$html)
        {
            return $html;
        }

        $doc = new DOMDocument();
        libxml_use_internal_errors(true);
        $readyContent = $this->beforeLoadHtml($html);
        $doc->loadHTML('<?xml encoding="utf-8"?>' . $readyContent);
        libxml_clear_errors();
        $doc->encoding = 'utf-8';
        $doc->formatOutput = true;

        $doc->removeChild($doc->firstChild); // remove html tag
        $doc->removeChild($doc->firstChild); // remove xml fix
        // make wip tag content as first child
        $doc->replaceChild(
            $doc->firstChild->firstChild->firstChild,
            $doc->firstChild
        );

        $html = $doc->saveHTML($doc->documentElement);
        $html = $this->afterSaveHtml($html);

        return $html;
    }

    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    protected function beforeLoadHtml(string $html): string
    {
        $html = "<wip>{$html}</wip>";
        $html = $this->fixNpTagsRegex($html);

        return $html;
    }

    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    protected function afterSaveHtml(string $html): string
    {
        $html = $this->fixNpTagsRegex($html, true);
        $html = \preg_replace('#^\s*<wip>(.*)</wip>\s*$#si', '$1', $html);

        return $html;
    }

    /**
     * @param string $html
     * @param bool   $revertMode
     * @return string
     */
    protected function fixNpTagsRegex(string $html, bool $revertMode = false): string
    {
        if (!$revertMode)
        {
            return \preg_replace(
                '#<(/?)(\w+):(\w+)( [^>]+)?>#i',
                '<$1$2-npfix-$3$4>',
                $html
            );
        }

        return \preg_replace(
            '#<(/?)(\w+)-npfix-(\w+)( [^>]+)?>#i',
            '<$1$2:$3$4>',
            $html
        );
    }
}
