<?php

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

abstract class AbstractTabbed extends AbstractTag
{
    /** @var int[] */
    public static $uniqueContainer = [];

    /** @var int[] */
    public static $uniqueTab = [];

    /**
     * @param \XF\App                                   $app
     * @param \XF\BbCode\Renderer\AbstractRenderer|null $renderer
     * @param string|null                               $rendererType
     */
    public function __construct(\XF\App $app, \XF\BbCode\Renderer\AbstractRenderer $renderer = null, $rendererType = null)
    {
        parent::__construct($app, $renderer, $rendererType);

        if (self::$uniqueContainer === [] || self::$uniqueTab === [])
        {
            self::$uniqueContainer[''] = \XF::$time;
            self::$uniqueTab[''] = \XF::$time;
        }
    }

    protected abstract function getIdContext(): string;

    protected function getEntityKeyFromOptions(array $renderOptions): string
    {
        $entity = $renderOptions['entity'] ?? null;
        if ($entity instanceof \XF\Mvc\Entity\Entity)
        {
            $contentType = $entity->getEntityContentType();
            $contentId = $entity->getEntityId();
            if ($contentType !== null && $contentId !== null)
            {
                return $contentType . '_' . $contentId;
            }
        }

        return '';
    }

    protected function getUniqueContainerId(array $renderOptions): string
    {
        $key = $this->getEntityKeyFromOptions($renderOptions);
        if (!isset(self::$uniqueContainer[$key]))
        {
            self::$uniqueContainer[$key] = 0;
        }
        $id = (++self::$uniqueContainer[$key]);

        return 'bbcode_' . $this->getIdContext() . (\strlen($key) !== 0 ? $key . '_' : '') . $id;
    }

    protected function getUniqueTabId(string $containerId, array $renderOptions): string
    {
        $key = $this->getEntityKeyFromOptions($renderOptions);
        if (!isset(self::$uniqueTab[$key]))
        {
            self::$uniqueTab[$key] = 0;
        }
        $id = (++self::$uniqueTab[$key]);

        return $containerId . '_tab_' . $id;
    }

    /**
     * @param string   $containerId
     * @param array    $tagChildren
     * @param string[] $tagOptions
     * @param array    $tag
     * @param array    $renderOptions
     * @return array
     * @noinspection PhpUnusedParameterInspection
     */
    protected function getSlides(string $containerId, array $tagChildren, array $tagOptions, array $tag, array $renderOptions): array
    {
        list($wipHeaders, $wipPanes) = $this->getHeadersAndPanes($tagChildren, $renderOptions, ['slide', 'slide_header']);

        $slides = [];
        foreach ($wipPanes as $tabIndex => $child)
        {
            $id = $this->getUniqueTabId($containerId, $renderOptions);
            $title = $wipHeaders[$tabIndex] ?? '';
            $sliderOption = $child['option'];
            $slides[] = [
                'id'      => $id,
                'content' => $child['content'],
                'options' => $this->parseSlideOptions($id, $title, \is_array($sliderOption) ? $sliderOption : [$sliderOption], $renderOptions, $tagOptions)
            ];
        }

        return $slides;
    }

    /**
     * @param array    $tagChildren
     * @param array    $options
     * @param string[] $specialTags
     * @return array
     */
    protected function getHeadersAndPanes(array $tagChildren, array $options, array $specialTags): array
    {
        $options = \array_merge($options, [
            'tabBbCode'       => true,
            'tabHeaderBbCode' => true,
        ]);

        $newMode = true;
        $specialTagsProbe = \array_fill_keys($specialTags, true);
        foreach ($tagChildren as $tag)
        {
            if (\is_array($tag) && empty($specialTagsProbe[$tag['tag']]))
            {
                $newMode = false;
                break;
            }
        }
        $wip = [];
        if ($newMode)
        {
            foreach ($tagChildren as $tag)
            {
                if (\is_array($tag))
                {
                    $wip[] = [
                        'tag'     => $tag['tag'],
                        'content' => $this->renderEscapedSubTree($tag['children'], $options),
                        'option'  => $this->getTagOptions($tag['option'], $options),
                    ];
                }
            }
        }
        if (!$wip)
        {
            $content = $this->renderer->renderSubTree($tagChildren, $options);
            $raw = $this->getSpecialTags($content, $specialTags, $options);
            if (!$raw)
            {
                // try closing the slide tag
                $raw = $this->getSpecialTags($content . '{/slide}', $specialTags, $options);
            }

            $wip = [];
            foreach ($raw as &$val)
            {
                $slideOptions = $this->getTagOptions($val['option'], $options, true, '|', false);
                $slideOptions = \array_map(function ($fragment) {
                    $fragment = $this->tidyHTML($fragment);

                    return $this->templater->preEscaped($fragment);
                }, $slideOptions);
                $wip[] = [
                    'tag'     => $val['tag'],
                    'content' => $this->templater->preEscaped($val['content']),
                    'option'  => $slideOptions,
                ];
            }
        }

        // sort into headers & planes
        return $this->splitSpecialTags($wip);
    }

    protected function splitSpecialTags(array $wip): array
    {
        $wipHeaders = [];
        $wipTabs = [];
        $tabIndex = 0;
        foreach ($wip as $slide)
        {
            switch ($slide['tag'])
            {
                case 'slide':
                case 'tab':
                    $wipTabs[] = $slide;
                    $tabIndex = \count($wipTabs);
                    break;
                case 'slide_header': // not massively extendable...
                case 'tab_header':
                    $wipHeaders[$tabIndex] = $slide['content'];
                    $tabIndex += 1;
                    break;
            }
        }

        return [$wipHeaders, $wipTabs];
    }

    /**
     * @param string                $id
     * @param string|\XF\PreEscaped $title
     * @param string[]              $slideOptions
     * @param array                 $renderingOptions
     * @param string[]              $tagOptions
     * @return array
     */
    abstract function parseSlideOptions(string $id, $title, array $slideOptions, array $renderingOptions, array $tagOptions): array;

    /**
     * @param string[]              $tagOptions
     * @param string|\XF\PreEscaped $title
     * @return array
     */
    abstract protected function getDefaultSlideOptions(array $tagOptions, $title): array;
}