<?php

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

use SV\AdvancedBbCodesPack\BbCode\Renderer\StyleTrait;
use XF\BbCode\Renderer\Html;

/**
 * Class AbstractTag
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
abstract class AbstractTag
{
    use StyleTrait;

    /**
     * @var \XF\App
     */
    protected $app;

    /**
     * @var \XF\BbCode\Renderer\AbstractRenderer
     */
    protected $renderer;

    /**
     * @var string
     */
    protected $rendererType;

    /**
     * @var \XF\Str\Formatter
     */
    protected $formatter;

    /**
     * @var \XF\Template\Templater
     */
    protected $templater;

    /**
     * @param \XF\App                                   $app
     * @param \XF\BbCode\Renderer\AbstractRenderer|null $renderer
     * @param string|null                               $rendererType
     */
    public function __construct(\XF\App $app, \XF\BbCode\Renderer\AbstractRenderer $renderer = null, string $rendererType = null)
    {
        $this->app = $app;
        $this->renderer = $renderer;
        $this->formatter = $app->stringFormatter();
        $this->templater = $renderer instanceof Html ? $renderer->getTemplater() : $app->templater();
        $this->rendererType = $rendererType;
    }

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     */
    abstract public function render(array $tagChildren, $tagOption, array $tag, array $options);

    /**
     * @param array                                $tagChildren
     * @param string|string[]                      $tagOption
     * @param array                                $tag
     * @param array                                $options
     * @param \XF\BbCode\Renderer\AbstractRenderer $renderer
     * @return string|\XF\PreEscaped
     */
    public static function renderStatic(array $tagChildren, $tagOption, array $tag, array $options, \XF\BbCode\Renderer\AbstractRenderer $renderer)
    {
        // if this function is called, it implies the add-on is disabled
        static $addonEnabled = null;
        if ($addonEnabled === null)
        {
            $addonsCache = \XF::app()->container('addon.cache');
            $addonEnabled = isset($addonsCache['SV/SignupAbuseBlocking']);
        }
        if (!$addonEnabled)
        {
            $renderer->renderUnparsedTag($tag, $options);
        }

        $tagRenderer = new static(\XF::app(), $renderer);

        return $tagRenderer->render($tagChildren, $tagOption, $tag, $options);
    }

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     * @noinspection PhpUnusedParameterInspection
     */
    protected function renderEditorInternal(array $tagChildren, $tagOption, array $tag, array $options)
    {
        return nl2br(
            $this->renderer->renderString($tag['original'][0], $options)
            . $this->renderer->renderSubTreePlain($tagChildren)
            . $this->renderer->renderString($tag['original'][1], $options)
        );
    }

    /**
     * @param array $tagChildren
     * @param array $options
     * @return string|\XF\PreEscaped
     */
    protected function renderEscapedSubTree(array $tagChildren, array $options)
    {
        return $this->templater->preEscaped($this->renderer->renderSubTree($tagChildren, $options));
    }

    /**
     * @param string $content
     * @param array  $tags
     * @param array  $options
     * @param bool   $getContentBetweenTags
     * @return array
     * @noinspection PhpUnusedParameterInspection
     * @noinspection RegExpUnnecessaryNonCapturingGroup
     * @noinspection RegExpRedundantEscape
     */
    protected function getSpecialTags(string $content, array $tags, array $options, bool $getContentBetweenTags = false): array
    {
        $tags = \implode('|', $tags);
        $afterTagBlank = '\s|&\#8203;';

        $count = \preg_match_all(
            '#(?P<beforeTag>(?:<[^>]+?>)+)?{(?P<tag>' . $tags . ')(=(?P<option>\[([\w\d]+)(?:=.+?)?\].+?\[\/\5\]|[^{}]+)+?)?}' .
            '(?P<afterTag>(?:(?:' . $afterTagBlank . ')*<\/[^>]+?>(?:' . $afterTagBlank . ')*)+)?(?P<content>.*?)' .
            '{\/\2}(?!(?:\W+)?{\/\2})(?P<outside>.*?)(?={(?:' . $tags . ')(?:=[^}]*)?}|$)#msi',
            $content,
            $matches,
            PREG_SET_ORDER
        );

        // Prevent html breaks
        for (; $count > 0; $count--)
        {
            $k = $count - 1;
            $content = $matches[$k]['content'];

            // Restore in a clean way a messy code between normal & special tags
            $beforeTag = $matches[$k]['beforeTag'] ?? '';
            $afterTag = $matches[$k]['afterTag'] ?? '';
            if (\strlen($beforeTag) && \strlen($afterTag))
            {
                $htmlBeforeTagData = \explode('>', $beforeTag);
                $htmlBeforeTagData = \array_filter($htmlBeforeTagData);
                $htmlBeforeTagData = \array_values($htmlBeforeTagData);

                $htmlAfterTagData = \explode('>', $afterTag);
                $htmlAfterTagData = \array_filter($htmlAfterTagData);
                $htmlAfterTagData = \array_reverse($htmlAfterTagData);

                $catchupContent = '';

                for ($i = 0; ; $i++)
                {
                    if (!isset($htmlBeforeTagData[$i], $htmlAfterTagData[$i]))
                    {
                        break;
                    }

                    $htmlBeforeTagLimitPos = \strpos($htmlBeforeTagData[$i], ' ');
                    if ($htmlBeforeTagLimitPos !== false)
                    {
                        $htmlBeforeTag = \substr($htmlBeforeTagData[$i], 1, $htmlBeforeTagLimitPos - 1);
                    }
                    else
                    {
                        //ie: b>
                        $htmlBeforeTag = \substr($htmlBeforeTagData[$i], -1);
                    }

                    $htmlAfterTagStartPos = \strpos($htmlAfterTagData[$i], '</');
                    if ($htmlAfterTagStartPos !== false)
                    {
                        $htmlAfterTag = \substr($htmlAfterTagData[$i], $htmlAfterTagStartPos + 2, \strlen($htmlAfterTagData[$i]));
                    }
                    else
                    {
                        break;
                    }

                    if ($htmlBeforeTag === $htmlAfterTag)
                    {
                        $catchupContent .= $htmlBeforeTagData[$i] . '>';
                    }
                }

                if ($catchupContent)
                {
                    $content = $catchupContent . $content;
                }
            }

            // Between special tags management
            $extraData = $matches[$k]['outside'];
            $extraDataCheck = \str_replace('<br />', '', $extraData);
            $extraDataCheck = \trim($extraDataCheck);

            if (empty($extraDataCheck))
            {
                $content = $this->tidyHTML($content);
                $matches[$k]['content'] = $content;

                continue;
            }

            if (!$getContentBetweenTags)
            {
                $extraData = $extraDataCheck;
                $contentToErase = \preg_split('#\s*<?[^>]+?>\s*#', $extraData);

                foreach ($contentToErase as $text)
                {
                    if ($text)
                    {
                        $extraData = \str_replace($text, '', $extraData);
                    }
                }
            }

            $content .= $extraData;
            $content = $this->tidyHTML($content);

            $matches[$k]['content'] = $content;
        }

        return $matches;
    }

    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    protected function tidyHTML(string $html): string
    {
        if (!\utf8_trim($html))
        {
            return $html;
        }

        $doc = new \DOMDocument();
        libxml_use_internal_errors(true);
        $readyContent = $this->beforeLoadHtml($html);
        $doc->loadHTML('<?xml encoding="utf-8"?>' . $readyContent);
        libxml_clear_errors();
        $doc->encoding = 'utf-8';
        $doc->formatOutput = true;

        $doc->removeChild($doc->firstChild); // remove html tag
        $doc->removeChild($doc->firstChild); // remove xml fix
        // make wip tag content as first child
        $doc->replaceChild(
            $doc->firstChild->firstChild->firstChild,
            $doc->firstChild
        );

        $html = $doc->saveHTML($doc->documentElement);
        $html = $this->afterSaveHtml($html);

        return $html;
    }

    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    protected function beforeLoadHtml(string $html): string
    {
        $html = "<wip>{$html}</wip>";
        $html = $this->fixNpTagsRegex($html);

        return $html;
    }

    /** @noinspection PhpUnnecessaryLocalVariableInspection */
    protected function afterSaveHtml(string $html): string
    {
        $html = $this->fixNpTagsRegex($html, true);
        $html = \preg_replace('#^\s*<wip>(.*)</wip>\s*$#si', '$1', $html);

        return $html;
    }

    protected function fixNpTagsRegex(string $html, bool $revertMode = false): string
    {
        if (!$revertMode)
        {
            return \preg_replace(
                '#<(/?)(\w+):(\w+)( [^>]+)?>#i',
                '<$1$2-npfix-$3$4>',
                $html
            );
        }

        return \preg_replace(
            '#<(/?)(\w+)-npfix-(\w+)( [^>]+)?>#i',
            '<$1$2:$3$4>',
            $html
        );
    }

    /**
     * See XF\BbCode\RuleSet::OPTION_KEYS_* for possible values.
     * const OPTION_KEYS_NO = 0;
     * const OPTION_KEYS_ONLY = 1;
     * const OPTION_KEYS_BOTH = 2;
     *
     * @return int
     */
    public function hasNativeMultipleOptionSupport(): int
    {
        return \XF\BbCode\RuleSet::OPTION_KEYS_BOTH;
    }

    /**
     * @param null|string|string[] $tagOption
     * @param string               $implode
     * @return string
     */
    protected function getTagOptionAsString($tagOption, string $implode = ' '): string
    {
        if ($tagOption === null)
        {
            return '';
        }

        $s = \is_array($tagOption) ? \implode($implode, $tagOption) : (string)$tagOption;

        if (strlen($s) !== 0)
        {
            $s = \utf8_trim($s);
        }

        return $s;
    }

    /**
     * @param null|string|array $tagOption
     * @param array        $options
     * @param bool         $multiOptionSupport
     * @param string       $separator
     * @param bool         $renderString
     * @return string[]
     */
    protected function getTagOptions($tagOption, array $options, bool $multiOptionSupport = true, string $separator = '|', bool $renderString = true): array
    {
        if ($tagOption === null)
        {
            return [];
        }

        if (\is_array($tagOption))
        {
            return $tagOption;
        }

        if (!$multiOptionSupport)
        {
            return [$tagOption];
        }

        if ($renderString)
        {
            $options['stopLineBreakConversion'] = 1 + (int)($options['stopLineBreakConversion'] ?? 0);
            $options['stopSmilies'] = 1 + (int)($options['stopSmilies'] ?? 0);
            $tagOption = (string)$this->renderer->renderString($tagOption, $options);
        }

        $tagOptionsRaw = \explode($separator, $tagOption) ?: [];

        $tagOptions = [];
        foreach ($tagOptionsRaw as $value)
        {
            if (\preg_match("#^\s*([a-zA-Z_-]+)\s*:\s*(.*)\s*$#us", $value, $matches))
            {
                if ($matches[2] === '' || \is_integer($matches[1]))
                {
                    $tagOptions[] = $matches[1];
                }
                else
                {
                    // preserver the raw value
                    $tagOptions[] = $value;

                    $tmp = \strtolower($matches[1]);
                    // skip raw links
                    /** @noinspection PhpConditionAlreadyCheckedInspection */
                    if (\substr($matches[2], 0, 2) === '//' && ($tmp !== 'http' || $tmp !== 'https'))
                    {
                        // raw link, reconstruct
                        continue;
                    }

                    // normalize key names
                    $key = \str_replace('-', '_', $matches[1]);
                    $tagOptions[$key] = $matches[2];
                }
            }
            else if ($value !== '')
            {
                $tagOptions[] = $value;
            }
        }

        return $tagOptions;
    }

    /**
     * @param string[] $tagOptions
     * @return string[]
     */
    protected function toLowerCaseTagOptions(array $tagOptions): array
    {
        $caseInsensitiveTagOptions = [];
        foreach ($tagOptions as $key => $value)
        {
            if (\is_integer($key))
            {
                $value = \strtolower($value);
            }
            else
            {
                $key = \strtolower($key);
            }
            $caseInsensitiveTagOptions[$key] = $value;
        }

        return $caseInsensitiveTagOptions;
    }

    public function getDefaultOptions(): array
    {
        return [];
    }

    protected function getOptionConstraints(): array
    {
        return [
            'defaultWidthType'  => 'px',
            'defaultHeightType' => 'px',
            'allowBlock'        => true,
            'allowFloat'        => true,
            'allowWidth'        => true,
            'allowAutoWidth'    => true,
            'allowHeight'       => false,
            'allowAutoHeight'   => true,
        ];
    }

    /**
     * @param string[] $tagOptions
     * @param string[] $renderingOptions
     * @return string[]
     */
    protected function parseTagOptions(array $tagOptions, array $renderingOptions): array
    {
        $options = $this->getDefaultOptions();
        $constraints = $this->getOptionConstraints();

        $lowerTagOptions = $this->toLowerCaseTagOptions($tagOptions);
        $this->parseBlockAlignOption($lowerTagOptions, $options, $constraints);
        $this->parseDimensionOptions($lowerTagOptions, $options, $constraints);
        $this->parseUniqueTagOptions($lowerTagOptions, $tagOptions, $options, $renderingOptions, $constraints);

        return $options;
    }

    protected function parseUniqueTagOptions(array &$lowerTagOptions, array &$tagOptions, array &$options, array $renderingOptions, array $constraints)
    {
    }

    /**
     * @param string|array $option
     * @return string
     */
    protected function getImageAltText($option): string
    {
        if (\is_array($option))
        {
            if (isset($option['alt']))
            {
                return $option['alt'];
            }
            if (isset($option['title']))
            {
                return $option['title'];
            }
        }

        return '';
    }

    /**
     * @param string|array $option
     * @return bool
     */
    protected function isFullAttachView($option): bool
    {
        if ($option)
        {
            if (\is_array($option) && isset($option['type']) && \strtolower($option['type']) === 'full')
            {
                return true;
            }
            else if (\is_string($option) && \strtolower($option) === 'full')
            {
                return true;
            }
        }

        return false;
    }

    /**
     * @param int   $attachmentId
     * @param array $renderingOptions
     * @return array|null
     */
    protected function getAttachmentParams(int $attachmentId, array $renderingOptions)
    {
        if (empty($renderingOptions['attachments'][$attachmentId]))
        {
            return null;
        }

        /** @var \XF\Entity\Attachment $attachment */
        $attachment = $renderingOptions['attachments'][$attachmentId];

        $extension = \strtolower($attachment->getExtension());
        if (!\in_array($extension, ['gif', 'jpe', 'jpeg', 'jpg', 'png'], true))
        {
            return ['valid' => false];
        }

        $viewAttachments = $renderingOptions['viewAttachments'] ?? false;
        return [
            'valid'      => $viewAttachments,
            'id'         => $attachmentId,
            'attachment' => $attachment,
            'full'       => true,
            'canView'    => $viewAttachments,
        ];
    }

    /**
     * @param string[] $tagOptions
     * @param string[] $options
     * @param string[] $constraints
     */
    protected function parseDimensionOptions(array &$tagOptions, array &$options, array $constraints)
    {
        $allowWidth = (bool)($constraints['allowWidth'] ?? false);
        $allowHeight = (bool)($constraints['allowHeight'] ?? false);
        if (!$allowWidth && !$allowHeight)
        {
            return;
        }

        $defaultWidthType = $constraints['defaultWidthType'] ?? 'px';
        $defaultHeightType = $constraints['defaultHeightType'] ?? 'px';
        $allowAutoWidth = (bool)($constraints['allowAutoWidth'] ?? false);
        $allowAutoHeight = (bool)($constraints['allowAutoHeight'] ?? false);
        $allowPercentageHeight = (bool)($constraints['allowPercentageHeight'] ?? false);

        $width = $options['width'] ?? null;
        $widthType = $options['width_type'] ?? null;
        $height = $options['height'] ?? null;
        $heightType = $options['height_type'] ?? null;

        $heightMatch = $allowPercentageHeight
            ? '/^(\d+)(|px|%)$/i'
            : '/^(\d+)(|px)$/i';

        foreach ($tagOptions as $index => $option)
        {
            if (!\is_integer($index))
            {
                switch ($index)
                {
                    case 'width':
                        if ($allowAutoWidth && \strcasecmp($option, 'auto') === 0)
                        {
                            $width = 'auto';
                            $widthType = 'auto';
                            unset($tagOptions[$index]);
                        }
                        else if (\preg_match('/^(\d+)(|px|%)$/i', $option, $matches))
                        {
                            $width = $matches[1];
                            $widthType = $matches[2] ?: $defaultWidthType;
                            unset($tagOptions[$index]);
                        }
                        break;
                    case 'height':
                        if ($allowAutoHeight && \strcasecmp($option, 'auto') === 0)
                        {
                            $height = 'auto';
                            $heightType = 'auto';
                            unset($tagOptions[$index]);
                        }
                        else if (\preg_match($heightMatch, $option, $matches))
                        {
                            $height = $matches[1];
                            $heightType = $matches[2] ?: $defaultHeightType;
                            unset($tagOptions[$index]);
                        }
                        break;
                }
                continue;
            }

            // width & height no support for pixel count for XF1 compatibility :(
            if (\preg_match('/^(\d+)(|px|%)(?:x(\d+))?$/i', $option, $matches))
            {
                $width = $matches[1];
                $widthType = $matches[2] ?: $defaultWidthType;
                $heightType = 'px';
                if (isset($matches[3]))
                {
                    $height = $matches[3];
                }
                //if (isset($matches[4]))
                //{
                //    $heightType = $matches[4];
                //}
                unset($tagOptions[$index]);
                continue;
            }

            // width + height -
            if (\preg_match('/^(\d+)(|px|%)x(\d+)$/', $option, $matches))
            {
                $width = $matches[1];
                $widthType = $matches[2] ?: $defaultWidthType;
                $height = $matches[3];
                $heightType = 'px';
                unset($tagOptions[$index]);
            }
        }

        if ($allowWidth)
        {
            if ($width === 'auto')
            {
                $widthType = $allowAutoWidth ? 'auto' : $defaultWidthType;
            }
            else if (!\strlen($widthType))
            {
                $widthType = $defaultWidthType;
            }

            switch ($widthType)
            {
                case 'px':
                    $maxWidth = $constraints['max_width_px'] ?? ($constraints['max_width'] ?? null);
                    if ($maxWidth && $width > $maxWidth)
                    {
                        $width = $maxWidth;
                    }
                    break;

                case '%':
                    if ($width > 100)
                    {
                        $width = 100;
                    }
                    $maxWidth = $constraints['max_width_percent'] ?? null;
                    if ($maxWidth && $width > $maxWidth)
                    {
                        $width = $maxWidth;
                    }
                    break;

                case 'auto':
                    if ($width === null || $width === 'auto')
                    {
                        $width = null;
                        $widthType = null;
                    }
                    break;
            }

            $options['width'] = $width;
            $options['width_type'] = $widthType;
        }

        if ($allowHeight)
        {
            if ($height === 'auto')
            {
                $heightType = null;
            }
            else if ($height !== null)
            {
                $heightType = $heightType ?: $defaultHeightType;
                $maxHeight = $constraints['max_height'] ?? null;
                if ($maxHeight && $height > $maxHeight)
                {
                    $height = $maxHeight;
                }
            }
            else
            {
                // no height effectively means auto
                $height = null;
                $heightType = null;
            }

            $options['height'] = $height;
            $options['height_type'] = $heightType;
        }
    }

    /**
     * @param string[] $tagOptions
     * @param string[] $options
     * @param string[] $constraints
     */
    protected function parseBlockAlignOption(array &$tagOptions, array &$options, array $constraints)
    {
        $allowBlock = (bool)($constraints['allowBlock'] ?? false);
        $allowFloat = (bool)($constraints['allowFloat'] ?? false);
        if (!$allowBlock && !$allowFloat)
        {
            return;
        }

        foreach ($tagOptions as $index => $option)
        {
            if (!\is_integer($index))
            {
                switch ($index)
                {
                    case 'block_align':
                        unset($tagOptions[$index]);
                        break;
                    default:
                        continue 2;
                }
            }

            switch ($option)
            {
                case 'bleft':
                case 'bcenter':
                case 'bright':
                    if ($allowBlock)
                    {
                        $options['block_align'] = $option;
                    }
                    unset($tagOptions[$index]);
                    continue 2;

                case 'fleft':
                case 'fright':
                    if ($allowFloat)
                    {
                        $options['block_align'] = $option;
                        unset($tagOptions[$index]);
                    }
                    unset($tagOptions[$index]);
                    continue 2;
            }
        }
    }

    /**
     * @param string   $option
     * @param string[] $options
     * @param bool     $includeVertical
     * @return bool
     */
    protected function parseTitleAlignmentOptions(string $option, array &$options, bool $includeVertical = false): bool
    {
        switch ($option)
        {
            case 'left':
            case 'center':
            case 'right':
                $options['title_align'] = $option;

                return true;

            case 'top':
            case 'bottom':
                if ($includeVertical)
                {
                    $options['title_align_vertical'] = $option;
                }

                return true;

            default:
                return false;
        }
    }

    protected function stringyFloat(string $option): float
    {
        /** @noinspection PhpWrongStringConcatenationInspection */
        return \strval(\floatval(\trim($option))) + 0;
    }

    /**
     * @param string $option
     * @return bool
     */
    protected function stringyTruth(string $option): bool
    {
        if ($option === 'true')
        {
            return true;
        }
        else if ($option === 'false')
        {
            return true;
        }

        return (bool)$option;
    }

    protected function isUrl(string $string): bool
    {
        $isUrl = $this->bindToRenderer(function ($string) {
            /** @noinspection PhpUndefinedFieldInspection */
            return (\preg_match($this->allowedUrlProtocolRegex, $string) === 1);
        });

        return $isUrl($string);
    }

    /**
     * @param string $url
     * @param array  $options
     * @return array|null
     */
    protected function getUrl(string $url, array $options)
    {
        $href = $this->getValidUrl($url);
        if (!$href)
        {
            return null;
        }

        $href = $this->formatter->censorText($href);

        $attributes = $this->getLinkAttributes($href, $options);
        if (!$attributes)
        {
            return null;
        }

        return [
            'href'       => $href,
            'attributes' => $attributes
        ];
    }

    /**
     * @param string $url
     * @return string|bool
     */
    protected function getValidUrl(string $url)
    {
        if (!$url || !$this->isUrl($url))
        {
            return false;
        }

        $getValidUrl = $this->bindToRenderer(function ($url) {
            return $this->getValidUrl($url);
        });

        return $getValidUrl($url);
    }


    /**
     * @param string $url
     * @param array  $options
     * @return string|\XF\PreEscaped
     */
    protected function getLinkAttributes(string $url, array $options)
    {
        $link = $this->getRenderedLink('', $url, $options);

        if (!\preg_match('/<a href=".+" (.+)>/isU', $link, $matches))
        {
            return '';
        }

        return $matches[1];
    }

    /**
     * @param string $text
     * @param string $url
     * @param array  $options
     * @return string|\XF\PreEscaped
     */
    protected function getRenderedLink(string $text, string $url, array $options)
    {
        $getRenderedLink = $this->bindToRenderer(
            function ($text, $url, array $options) {
                return $this->getRenderedLink($text, $url, $options);
            }
        );

        return $getRenderedLink($text, $url, $options);
    }

    /**
     * @param array $children
     */
    protected function trimChildrenList(array &$children)
    {
        if (!$children)
        {
            return;
        }
        $trimChildrenList = $this->bindToRenderer(function (&$children) {
            $this->trimChildrenList($children);
        });
        $trimChildrenList($children);
    }

    /**
     * @param \Closure $closure
     * @return \Closure
     * @throws \LogicException
     */
    protected function bindToRenderer(\Closure $closure): \Closure
    {
        $bound = \Closure::bind($closure, $this->renderer, \get_class($this->renderer));

        if (!$bound)
        {
            throw new \LogicException('Could not bind to renderer');
        }

        return $bound;
    }
}