<?php

/*
 * This file is part of a XenForo add-on.
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

/**
 * A BB code renderer for [Anchor] tags.
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
class Anchor extends AbstractTag
{
    use TagWithSingleOptionTrait;

    protected $anchorPrefix = 'message-anchor';

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     */
    public function render(array $tagChildren, $tagOption, array $tag, array $options)
    {
        $this->trimChildrenList($tagChildren);
        $content = $this->renderEscapedSubTree($tagChildren, $options);

        $tagOption = $this->getTagOptionAsString($tagOption);
        // Anchor point
        if (strlen($tagOption) === 0)
        {
            $anchor = \utf8_strtolower(\htmlspecialchars($content));
            return "<a id='{$this->anchorPrefix}-{$anchor}' class='bb-code--anchor'></a>";
        }

        // Anchor link
        $anchorSuffix = \utf8_strtolower(\htmlspecialchars($tagOption));

        if ($anchorSuffix[0] === '#')
        {
            $anchorSuffix = \substr($anchorSuffix, 1);
        }
        $anchor = "#{$this->anchorPrefix}-{$anchorSuffix}";

        $url = $anchor;
        if ($this->app->options()->svFullAnchorUrl ?? false)
        {
            $entity = $options['entity'] ?? null;
            if ($entity instanceof \XF\Mvc\Entity\Entity)
            {
                if (!\is_callable([$entity, 'canView']) || $entity->canView())
                {
                    $url = $this->getEntityLink($entity, $anchor);
                }
            }
        }

        $options['noFollowUrl'] = true;
        $options['noProxy'] = true;
        return $this->getRenderedLink($content, $url, $options);
    }

    protected function getEntityLink(\XF\Mvc\Entity\Entity $entity, string $anchor): string
    {
        $router = $this->app->router('public');
        $options = $this->app->options();
        if ($entity instanceof \XF\Entity\Post)
        {
            if ($entity->Thread)
            {
                $page = \floor($entity->position / $options->messagesPerPage) + 1;

                return $router->buildLink('threads', $entity->Thread, ['page' => $page]) . $anchor;
            }
        }
        else if ($entity instanceof \XF\Entity\ConversationMessage)
        {
            if ($entity->Conversation)
            {
                /** @var \XF\Repository\ConversationMessage $convMessageRepo */
                $convMessageRepo = $this->app->repository('XF:ConversationMessage');
                $messagesBefore = $convMessageRepo->findEarlierMessages($entity->Conversation, $entity)->total();
                $page = \floor($messagesBefore / $options->messagesPerPage) + 1;

                return $router->buildLink('conversations', $entity->Conversation, ['page' => $page]) . $anchor;
            }
        }
        else if ($entity instanceof \XF\Entity\ProfilePost)
        {
            /** @var \XF\Repository\ProfilePost $profilePostRepo */
            $profilePostRepo = $this->app->repository('XF:ProfilePost');

            $profilePostFinder = $profilePostRepo->findProfilePostsOnProfile($entity->ProfileUser);
            $profilePostsCount = $profilePostFinder->where('post_date', '>', $entity->post_date)->total();

            $page = \floor($profilePostsCount / $options->messagesPerPage) + 1;

            return $router->buildLink('members', $entity->ProfileUser, ['page' => $page]) . $anchor;
        }
        else if ($entity instanceof \XF\Entity\ProfilePostComment)
        {
            $profilePost = $entity->ProfilePost;
            if ($profilePost)
            {
                /** @var \XF\Repository\ProfilePost $profilePostRepo */
                $profilePostRepo = $this->app->repository('XF:ProfilePost');

                $profilePostFinder = $profilePostRepo->findProfilePostsOnProfile($profilePost->ProfileUser);
                $profilePostsCount = $profilePostFinder->where('post_date', '>', $profilePost->post_date)->total();

                $page = \floor($profilePostsCount / $options->messagesPerPage) + 1;

                $commentId = $entity->profile_post_comment_id;
                if (!isset($profilePost->latest_comment_ids[$commentId]))
                {
                    $anchor = '#profile-post-' . $profilePost->profile_post_id;
                }

                return $router->buildLink('members', $profilePost->ProfileUser, ['page' => $page]) . $anchor;
            }
        }
        else if ($entity instanceof \XF\Entity\ReportComment)
        {
            if ($entity->Report)
            {
                return $router->buildLink('reports', $entity->Report);
            }
        }

        return $anchor;
    }
}