<?php

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

/**
 * Class Bimg
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
class Bimg extends AbstractTag
{
    /**
     * @var int
     */
    const MAX_IMAGES = 2;

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     */
    public function render(array $tagChildren, $tagOption, array $tag, array $options)
    {
        $tagOptions = $this->getTagOptions($tagOption, $options);
        $tagOptions['images'] = $this->getImages($tagChildren, $options, static::MAX_IMAGES);
        $tagOptions = $this->parseTagOptions($tagOptions, $options);
        if (!$tagOptions['images'])
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }
        $template = $this->rendererType === 'emailHtml' ? 'email:sv_bbcode_bimg' : 'public:sv_bbcode_bimg';

        $result = $this->templater->renderTemplate($template, [
            'images'  => $tagOptions['images'],
            'options' => $tagOptions
        ]);

        // ensure empty title/alt tags aren't generated by XF as Chrome is buggy with them
        $result2 = \preg_replace(['#title="\\s*"#', '#alt="\\s*"#'], '', $result);
        if ($result2)
        {
            $result = $result2;
        }

        return $result;
    }

    /**
     * @param array $tagChildren
     * @param array $renderingOptions
     * @param int   $limit
     * @return array|null
     */
    protected function getImages(array $tagChildren, array $renderingOptions, int $limit = 0)
    {
        $images = [];

        // check for direct attachments & images as children. Requires children nodes to be parsed
        $this->trimChildrenList($tagChildren);
        foreach ($tagChildren as $key => $childTag)
        {
            if (!\is_array($childTag))
            {
                continue;
            }
            switch ($childTag['tag'])
            {
                case 'attach':
                    $childTagOptions = $this->getTagOptions($childTag['option'], $renderingOptions);
                    $attachmentId = (int)$this->renderer->renderSubTreePlain($childTag['children']);
                    $attachmentParams = $this->getAttachmentParams($attachmentId, $renderingOptions);
                    if ($attachmentParams !== null && \count($images) < static::MAX_IMAGES)
                    {
                        /** @var \XF\Entity\Attachment $attachment */
                        $attachment = $attachmentParams['attachment'];
                        $attachmentParams['full'] = $this->isFullAttachView($childTagOptions);
                        //$attachmentParams['alignClass'] = $this->getAttachAlignClass($option);
                        //$attachmentParams['styleAttr'] = $this->getAttachStyleAttr($option);
                        $attachmentParams['alt'] = $this->getImageAltText($childTagOptions) ?: ($attachment ? $attachment->filename : '');
                        $images[] = $attachmentParams;
                    }
                    unset($tagChildren[$key]);
                    break;
                case 'img':
                case 'bimg':
                    $url = $this->renderer->renderSubTreePlain($childTag['children']);
                    $image = $this->getImage(\utf8_trim($url), $renderingOptions);
                    if ($image)
                    {
                        $images[] = $image;
                    }
                    unset($tagChildren[$key]);
                    break;
                default:
                    return [];
            }
        }

        // check for raw URL links
        $content = $this->renderer->renderSubTreePlain($tagChildren);

        /** @noinspection RegExpSingleCharAlternation */
        $urls = \preg_split("/\||\n/u", $content, -1, PREG_SPLIT_NO_EMPTY);
        foreach ($urls as $url)
        {
            $image = $this->getImage(\utf8_trim($url), $renderingOptions);
            if (!$image)
            {
                return null;
            }

            $images[] = $image;
        }

        if ($limit)
        {
            $images = \array_slice($images, 0, $limit, false);
        }

        return $images;
    }

    /**
     * @param string $url
     * @param array  $renderingOptions
     * @return array|null
     */
    protected function getImage(string $url, array $renderingOptions)
    {
        $id = (int)$url;
        if($id)
        {
            $attachmentParams = $this->getAttachmentParams($id, $renderingOptions);
            if ($attachmentParams !== null)
            {
                return $attachmentParams;
            }
        }

        $dataUrl = $this->getValidUrl($url);
        if (!$dataUrl)
        {
            return null;
        }

        $censored = $this->formatter->censorText($dataUrl);
        if ($censored !== $dataUrl)
        {
            return null;
        }

        if ($renderingOptions['noProxy'])
        {
            $src = $dataUrl;
        }
        else
        {
            $linkInfo = $this->formatter->getLinkClassTarget($dataUrl);
            if ($linkInfo['local'])
            {
                $src = $dataUrl;
            }
            else
            {
                $src = $this->formatter->getProxiedUrlIfActive(
                    'image',
                    $dataUrl
                );
                if (!$src)
                {
                    $src = $dataUrl;
                }
            }
        }

        return [
            'alt'     => null,
            'src'     => $src,
            'dataUrl' => $dataUrl
        ];
    }

    /**
     * @param string $tagOption
     * @param array  $options
     * @return bool
     */
    protected function parseImgAlignmentOptions(string $tagOption, array &$options): bool
    {
        switch ($tagOption)
        {
            case 'diff-v':
            case 'diff-vertical':
                $options['comparison_orientation'] = 'vertical';
                return true;

            case 'diff-h':
            case 'diff-horizontal':
                $options['comparison_orientation'] = 'horizontal';
                return true;

            default:
                return false;
        }
    }

    protected function parseUniqueTagOptions(array &$lowerTagOptions, array &$tagOptions, array &$options, array $renderingOptions, array $constraints)
    {
        // unique options
        foreach ($lowerTagOptions as $key => $tagOption)
        {
            if (!\is_integer($key))
            {
                switch ($key)
                {
                    case 'url':
                        $url = $this->getUrl(\trim($tagOptions[$key]), $options);
                        if ($url)
                        {
                            $options['url'] = $url;
                            $options['lightbox'] = false;
                        }
                        break;
                    case 'title':
                        $options['title'] = $tagOption;
                        break;
                    case 'align':
                    case 'title_align':
                    case 'title_align_vertical':
                        $this->parseTitleAlignmentOptions(\strtolower($tagOption), $options, true);
                        break;
                    case 'image_align':
                        $this->parseImgAlignmentOptions(\strtolower($tagOption), $options);
                        break;
                    case 'title_inside':
                        $options['title_inside'] = $this->stringyTruth($tagOption);
                        break;
                    case 'lightbox':
                        $options['lightbox'] = $this->stringyTruth($tagOption);
                        break;
                    case 'no_lightbox':
                        $options['lightbox'] = !$this->stringyTruth($tagOption);
                        break;
                    case 'diff_pos':

                        $options['comparison_offset'] = $this->stringyFloat($tagOption);
                        break;
                }
                continue;
            }

            if ($this->parseTitleAlignmentOptions($tagOption, $options, true))
            {
                continue;
            }
            if ($this->parseImgAlignmentOptions($tagOption, $options))
            {
                continue;
            }

            switch ($tagOption)
            {
                case 'inside':
                    $options['title_inside'] = true;
                    continue 2;
                case 'no-lightbox':
                case 'no_lightbox':
                    $options['lightbox'] = false;
                    continue 2;
            }

            $cleaned = \trim($tagOptions[$key]);
            $url = $this->getUrl($cleaned, $options);
            if ($url)
            {
                $options['url'] = $url;
                $options['lightbox'] = false;
                continue;
            }

            // title
            if ($cleaned !== '')
            {
                $options['title'] = $cleaned;
            }
        }

        $options['images'] = $tagOptions['images'] ?? [];

        // comparison-mode
        if (\count($options['images']) > 1)
        {
            $options['comparison'] = true;
            $options['lightbox'] = false;
        }
    }

    public function getDefaultOptions(): array
    {
        $style = $this->getStyle();

        return [
            // global options
            'block_align'          => $style->getProperty('svBbcodeBimgDefaultBlockAlignment'),
            'width'                => $style->getProperty('svBbcodeBimgDefaultWidth'),
            'width_type'           => $style->getProperty('svBbcodeBimgDefaultWidthUnit'),

            // bimg options
            'title'                => '',
            'title_align'          => 'left',
            'title_align_vertical' => 'bottom',
            'title_inside'         => false,

            'comparison'             => false,
            'comparison_orientation' => 'horizontal',
            'comparison_offset'      => 0.5,

            'lightbox' => true,
            'url'      => ''
        ];
    }

    protected function getOptionConstraints(): array
    {
        return \array_merge(parent::getOptionConstraints(), [
            'max_width' => $this->getStyleProperty('svBbcodeBimgMaxWidth', 'svBbcodeBimgApplyWidthRestrictions')
        ]);
    }
}