<?php

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

/**
 * Class ModInterrupt
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
class ModInterrupt extends AbstractTag
{
    /**
     * @param string $context
     * @return bool|null
     */
    protected function validContext(string $context)
    {
        return $context === 'help' ? true : null;
    }

    /** @var array|null */
    protected $staffOnlyBbCode = null;

    protected function isStaffOnlyBbCode(string $tagName): bool
    {
        if ($this->staffOnlyBbCode === null)
        {
            $options = \XF::options();

            $modOnlyBbCode = $options->modOnlyBbCode ?? '';
            if (\strlen($modOnlyBbCode) && ($this->app instanceof \XF\Pub\App) && ($options->modOnlyBbCodeRender ?? false))
            {
                $tags = \explode(',', \str_replace("\n", ',', $modOnlyBbCode));
                $tags = \array_map('\trim', $tags);
                $tags = \array_filter($tags);
                $tags = \array_map('\utf8_strtolower', $tags);
                $this->staffOnlyBbCode = \array_fill_keys($tags, true);
            }
            else
            {
                $this->staffOnlyBbCode = [];
            }
        }

        if (isset($this->staffOnlyBbCode[$tagName]))
        {
            return true;
        }

        return false;
    }

    protected function populateRenderState(array $options, array &$extraCss, bool &$render = null, \XF\Entity\User &$user = null)
    {
        $helperRepo = \SV\StandardLib\Helper::repo();
        $entity = $options['entity'] ?? null;
        if ($entity instanceof \XF\Mvc\Entity\Entity)
        {
            /** @var \SV\Threadmarks\Entity\Threadmark $threadmark */
            if ($entity->isValidRelation('Threadmark') &&
                ($threadmark = $entity->getRelation('Threadmark')) &&
                ($threadmark->threadmark_category_id === (int)(\XF::options()->modInterruptThreadmarkCategory ?? 0)) &&
                $threadmark->isVisible())
            {
                $extraCss[] = 'mod_interrupt--staff-threadmark';
                $render = true;
            }
            if ($entity instanceof \XF\Entity\Page)
            {
                $extraCss[] = 'mod_interrupt--page';
                $render = true;
            }

            $user = $helperRepo->getUserEntity($entity);
            if (!$user && ($entity->isValidGetter('Content') || $entity->isValidRelation('Content')))
            {
                $user = $helperRepo->getUserEntity($entity->get('Content'));
            }
            // XFRM support, as it doesn't have a User/Content relationship...
            if (!$user && ($entity->isValidGetter('Resource') || $entity->isValidRelation('Resource')))
            {
                $user = $helperRepo->getUserEntity($entity->get('Resource'));
            }
        }
        else if ($this->renderer->getRules()->getSubContext() === 'preview')
        {
            if (isset($options['user']))
            {
                $user = $options['user'];
                $user = $helperRepo->getUserEntity($user);
            }
        }
    }

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     */
    public function render(array $tagChildren, $tagOption, array $tag, array $options)
    {
        $tagOptions = $this->getTagOptions($tagOption, $options);
        $tagOptions = $this->parseTagOptions($tagOptions, $options);

        $context = $this->renderer->getRules()->getContext();
        $render = $this->validContext($context);
        $tagName = $tag['tag'];
        $requireModerator = !$render && $this->isStaffOnlyBbCode($tagName);
        $extraCss = [];
        $user = null;
        if ($render === null)
        {
            $this->populateRenderState($options, $extraCss, $render, $user);
        }

        if ($user && $render !== false)
        {
            if ($user->is_moderator)
            {
                $extraCss[] = 'mod_interrupt--is-moderator';
                $render = true;
            }
            else if (!$requireModerator)
            {
                $render = true;
            }
        }

        if (!$render)
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }

        $content = $this->renderEscapedSubTree($tagChildren, $options);
        $template = $this->rendererType === 'emailHtml' ? 'email:sv_bbcode_quote_chunk' : 'public:sv_bbcode_modinterrupt';

        return $this->templater->renderTemplate($template, [
            'modType'   => $tagName,
            'content'   => $content,
            'extraCss'  => $extraCss ? \implode(' ', $extraCss) : '',
            'options'   => $tagOptions,
        ]);
    }

    protected function parseUniqueTagOptions(array &$lowerTagOptions, array &$tagOptions, array &$options, array $renderingOptions, array $constraints)
    {
        foreach ($lowerTagOptions as $key => $tagOption)
        {
            if (!\is_integer($key))
            {
                switch ($key)
                {
                    case 'title':
                        $options['title'] = $tagOption;
                        break;
                    //case 'title_align':
                    //    $this->parseTitleAlignmentOptions($tagOptions, $key, $option, $options);
                    //    break;
                }
                continue;
            }


            $cleaned = \trim($tagOption);
            // title
            if ($cleaned !== '')
            {
                $options['title'] = $cleaned;
            }
        }
    }

    public function getDefaultOptions(): array
    {
        return [
            'block_align' => 'left',
        ];
    }

    protected function getOptionConstraints(): array
    {
        // todo - requires css/html layout support for this
        return \array_merge(parent::getOptionConstraints(), [
            'allowBlock'  => false,
            'allowFloat'  => false,
            'allowWidth'  => false,
            'allowHeight' => false,
        ]);
    }
}