<?php

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

/**
 * Class Slider
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
class Slider extends AbstractTabbed
{
    protected function getIdContext(): string
    {
        return 'slider';
    }

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     */
    public function render(array $tagChildren, $tagOption, array $tag, array $options)
    {
        $tagOptions = $this->getTagOptions($tagOption, $options);
        $tagOptions = $this->parseTagOptions($tagOptions, $options);

        $containerId = $this->getUniqueContainerId($options);
        $slides = $this->getSlides($containerId, $tagChildren, $tagOptions, $tag, $options);
        if (!$slides)
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }
        $template = $this->rendererType === 'emailHtml' ? 'email:sv_bbcode_tabbed_content' : 'public:sv_bbcode_slider';

        return $this->templater->renderTemplate($template, [
            'options'     => $tagOptions,
            'slides'      => $slides,
            'containerId' => $containerId,
        ]);
    }

    protected function parseSliderInterval(array $constraints, string $value)
    {
        if (\preg_match('/^(\d+|\d\.\d)(ms|s)$/i', $value, $matches))
        {
            $interval = (int)$matches[0];
            $units = $matches[1];
            switch ($units)
            {
                case 's':
                    $interval = $interval * 1000;
                    break;
                case 'ms':
                    break;
                default:
                    return null;
            }
            if (($interval >= $constraints['interval_min']) && ($interval <= $constraints['interval_max']))
            {
                return $interval;
            }
        }

        return null;
    }

    protected function parseUniqueTagOptions(array &$lowerTagOptions, array &$tagOptions, array &$options, array $renderingOptions, array $constraints)
    {
        foreach ($lowerTagOptions as $key => $option)
        {
            if (!\is_integer($key))
            {
                switch ($key)
                {
                    case 'autoplay':
                        if ($constraints['allow_autoplay'])
                        {
                            $options['autoplay'] = $this->stringyTruth($option);
                        }
                        break;
                    case 'allow_touch':
                    case 'cmd':
                    case 'inside':
                    case 'noclick':
                    case 'num':
                        $options[$key] = $this->stringyTruth($option);
                        break;
                    case 'interval':
                        $interval = $this->parseSliderInterval($constraints, $option);
                        if ($interval !== null)
                        {
                            $options[$key] = $interval;
                        }
                        break;
                }
                continue;
            }

            switch ($option)
            {
                case 'autoplay':
                    if ($constraints['allow_autoplay'])
                    {
                        $options['autoplay'] = true;
                    }
                    continue 2;
                case 'allow_touch':
                case 'cmd':
                case 'inside':
                case 'noclick':
                case 'num':
                    $options[$option] = true;
                    continue 2;
            }

            $interval = $this->parseSliderInterval($constraints, $option);
            if ($interval !== null)
            {
                $options[$key] = $interval;
            }

            //$original = $tagOptions[$key];
            //$cleaned = \trim($original);
        }
    }

    public function getDefaultOptions(): array
    {
        $style = $this->getStyle();

        return [
            // global options
            'block_align' => $style->getProperty('svBbcodeSliderDefaultBlockAlignment'),
            'width'       => $style->getProperty('svBbcodeSliderDefaultWidth'),
            'width_type'  => $style->getProperty('svBbcodeSliderDefaultWidthUnit'),
            'height'      => $style->getProperty('svBbcodeSliderDefaultHeight'),

            // slider options
            'allow_touch' => $style->getProperty('svBbcodeSliderDefaultAllowTouch'),
            'autoplay'    => $style->getProperty('svBbcodeSliderDefaultAutoplay'),
            'interval'    => $style->getProperty('svBbcodeSliderDefaultInterval'),
            'cmd'         => false,
            'inside'      => false,
            'noclick'     => false,
            'num'         => false
        ];
    }

    protected function getOptionConstraints(): array
    {
        return \array_merge(parent::getOptionConstraints(), [
            'allowHeight' => true,
            'max_width'      => $this->getStyleProperty('svBbcodeSliderMaxWidth', 'svBbcodeSliderApplyWidthRestrictions'),
            'max_height'     => $this->getStyleProperty('svBbcodeSliderMaxHeight', 'svBbcodeSliderApplyHeightRestrictions'),
            'interval_min'   => $this->getStyleProperty('svBbcodeSliderMinInterval'),
            'interval_max'   => $this->getStyleProperty('svBbcodeSliderMaxInterval'),
            'allow_autoplay' => $this->getStyleProperty('svBbcodeSliderAllowAutoplay')
        ]);
    }

    /**
     * @param string                $id
     * @param string|\XF\PreEscaped $title
     * @param string[]              $slideOptions
     * @param array                 $renderingOptions
     * @param string[]              $tagOptions
     * @return array
     */
    function parseSlideOptions(string $id, $title, array $slideOptions, array $renderingOptions, array $tagOptions): array
    {
        $options = $this->getDefaultSlideOptions($tagOptions, $title);

        foreach ($slideOptions as $key => $slideOption)
        {
            if (!\is_integer($key))
            {
                switch ($key)
                {
                    case 'title':
                        $options['title'] = $slideOption;
                        break;
                    case 'open':
                        $options['open'] = $this->stringyTruth($slideOption);
                        break;
                    case 'align':
                    case 'title_align':
                        $this->parseTitleAlignmentOptions(\strtolower($slideOption), $options);
                        break;
                    case 'vertical_align':
                        $this->parseTitleAlignmentOptions(\strtolower($slideOption), $options, true);
                        break;
                    case 'height':
                        $options['height'] = $this->stringyFloat($slideOption);
                        break;
                    case 'full':
                        $options['attachment_full'] = $this->stringyTruth($slideOption);
                        break;
                    case 'attach':
                        $attachmentId = (int)$slideOption;

                        $attachmentParams = $this->getAttachmentParams($attachmentId, $renderingOptions);
                        if ($attachmentParams !== null)
                        {
                            $options['attachment_params'] = $attachmentParams;
                        }
                        break;
                }

                continue;
            }

            $original = $slideOption;
            $slideOption = \strtolower(\trim($slideOption));

            switch ($slideOption)
            {
                case 'left':
                case 'center':
                case 'right':
                    $options['title_align'] = $slideOption;
                    continue 2;

                case 'top':
                case 'bottom':
                    $options['title_align_vertical'] = $slideOption;
                    continue 2;

                case 'full':
                    $options['attachment_full'] = true;
                    continue 2;

                case 'open':
                    $options['open'] = true;
                    continue 2;
            }

            // #
            if (\preg_match('/^\d+$/', $slideOption))
            {
                $attachmentId = (int)$slideOption;

                $attachmentParams = $this->getAttachmentParams($attachmentId, $renderingOptions);
                if ($attachmentParams !== null)
                {
                    $options['attachment_params'] = $attachmentParams;
                    continue;
                }
            }

            // title
            if ($slideOption !== '')
            {
                $options['title'] = $original;
            }
        }

        if (!$options['attachment_params'])
        {
            $options['title_align_vertical'] = '';
            $options['attachment_full'] = false;
        }

        return $options;
    }

    /**
     * @param string[]              $tagOptions
     * @param string|\XF\PreEscaped $title
     * @return array
     */
    protected function getDefaultSlideOptions(array $tagOptions, $title): array
    {
        return [
            'title'                => $title,
            'title_align'          => 'left',
            'title_align_vertical' => '',
            'open'                 => false,
            'attachment_params'    => [],
            'attachment_full'      => false,
        ];
    }
}