<?php

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

/**
 * Class Tabs
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
class Tabs extends AbstractTabbed
{
    protected function getIdContext(): string
    {
        return 'tabs';
    }

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     */
    public function render(array $tagChildren, $tagOption, array $tag, array $options)
    {
        $tagOptions = $this->getTagOptions($tagOption, $options);
        $tagOptions = $this->parseTagOptions($tagOptions, $options);

        $containerId = $this->getUniqueContainerId($options);
        $slides = $this->getSlides($containerId, $tagChildren, $tagOptions, $tag, $options);
        $slides = $this->filterEmptyTabs($slides);
        if (!$slides)
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }
        $slides = $this->linkTabButtons($slides, $containerId);
        $template = $this->rendererType === 'emailHtml' ? 'email:sv_bbcode_tabbed_content' : 'public:sv_bbcode_tabs';

        foreach($slides as $slide)
        {
            if ($slide['options']['open'] ?? false)
            {
                $tagOptions['activeId'] = $slide['id'];
                break;
            }
        }

        if (!\array_key_exists('activeId', $tagOptions))
        {
            $tagOptions['activeId'] = \reset($slides)['id'];
        }

        return $this->templater->renderTemplate($template, [
            'activeId'    => $tagOptions['activeId'],
            'options'     => $tagOptions,
            'containerId' => $containerId,
            'tabs'        => $slides,
            'slides'      => $slides,
        ]);
    }

    /**
     * @param string                $id
     * @param string|\XF\PreEscaped $title
     * @param string[]              $slideOptions
     * @param array                 $renderingOptions
     * @param string[]              $tagOptions
     * @return array
     */
    function parseSlideOptions(string $id, $title, array $slideOptions, array $renderingOptions, array $tagOptions): array
    {
        $options = $this->getDefaultSlideOptions($tagOptions, $title);

        foreach ($slideOptions as $key => $slideOption)
        {
            if (!\is_integer($key))
            {
                switch ($key)
                {
                    case 'title':
                        $options['title'] = $slideOption;
                        break;
                    case 'open':
                        $options['open'] = $this->stringyTruth($slideOption);
                        break;
                    case 'align':
                    case 'title_align':
                        $this->parseTitleAlignmentOptions(\strtolower($slideOption), $options);
                        break;
                }

                continue;
            }

            $original = $slideOption;
            $slideOption = \strtolower(\trim($slideOption));

            switch ($slideOption)
            {
                case 'left':
                case 'center':
                case 'right':
                    $options['title_align'] = $slideOption;
                    continue 2;

                //case 'top':
                //case 'bottom':
                //    $options['title_align_vertical'] = $slideOption;
                //    continue 2;

                case 'open':
                    $options['open'] = true;
                    continue 2;
            }

            // title
            if ($slideOption !== '')
            {
                $options['title'] = $original;
            }
        }

        return $options;
    }

    /**
     * @param string[]              $tagOptions
     * @param string|\XF\PreEscaped $title
     * @return array
     */
    protected function getDefaultSlideOptions(array $tagOptions, $title): array
    {
        return [
            'title'                => $title,
            'title_align'          => 'left',
            'title_align_vertical' => '',
            'open'                 => false,
        ];
    }

    private function filterEmptyTabs(array $tabs): array
    {
        $ret = [];
        foreach ($tabs as $tab)
        {
            // if there is explicit title, skip filtering empty tabs
            $realTitle = \trim($tab['options']['title']);
            if ($realTitle !== '')
            {
                $ret[] = $tab;
                continue;
            }
            $realContent = \trim(\strip_tags($tab['content']));
            if ($realContent !== '' || \preg_match('/<img/', $tab['content']))
            {
                $ret[] = $tab;
            }
        }

        return $ret;
    }

    /**
     * @param array  $tabs
     * @param string $containerId
     * @return array
     */
    private function linkTabButtons(array $tabs, string $containerId): array
    {
        foreach ($tabs AS $internalId => $tab)
        {
            $content = $tab['content'];
            if (!\XF::isPreEscaped($content))
            {
                // only supported if is escaped to avoid XSS injection
                continue;
            }

            $content = (string)$content;
            $allMatches = [];
            /** @noinspection RegExpRedundantEscape */
            \preg_match_all('#{tab=(\d)}(.*?){\/tab}#i', $content, $matches, PREG_SET_ORDER);
            foreach($matches as $match)
            {
                $allMatches[$match[0]] = $match;
            }
            \preg_match_all('#\[tab=(\d)\](.*?)\[\/tab\]#i', $content, $matches, PREG_SET_ORDER);
            foreach($matches as $match)
            {
                $allMatches[$match[0]] = $match;
            }

            foreach ($allMatches AS $match)
            {
                $originalTabInternalId = (int)$match[1] - 1;
                $originalTab = $tabs[$originalTabInternalId] ?? null;
                if ($originalTab)
                {
                    $content = \str_replace($match[0], '<a class="sv-tabs-tab-link" data-xf-init="sv-tabs-tab-link" data-tab-id="' . $originalTab['id'] . '" data-tab-container-id="' . $containerId . '">' . $match[2] . '</a>', $content);
                }
            }

            $tabs[$internalId]['content'] = $this->templater->preEscaped($content);
        }

        return $tabs;
    }

    public function getDefaultOptions(): array
    {
        $style = $this->getStyle();

        return [
            'width'       => $style->getProperty('svBbcodeTabsDefaultWidth'),
            'width_type'  => $style->getProperty('svBbcodeTabsDefaultWidthUnit'),
            'block_align' => $style->getProperty('svBbcodeTabsDefaultBlockAlignment'),
            'title_align' => $style->getProperty('svBbcodeTabsDefaultTitleAlignment')
        ];
    }

    protected function getOptionConstraints(): array
    {
        return \array_merge(parent::getOptionConstraints(), [
            'allowHeight' => true,
            'max_width'   => $this->getStyleProperty('svBbcodeTabsMaxWidth', 'svBbcodeTabsApplyWidthRestrictions'),
            'max_height'  => $this->getStyleProperty('svBbcodeTabsMaxHeight', 'svBbcodeTabsApplyHeightRestrictions')
        ]);
    }
}