<?php

/*
 * This file is part of a XenForo add-on.
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */

namespace SV\AdvancedBbCodesPack\BbCode\Tag;

/**
 * A BB code renderer for [TIME] tags.
 *
 * @package SV\AdvancedBbCodesPack\BbCode\Tag
 */
class Time extends AbstractTag
{
    /**
     * @var array
     */
    protected $typeMap = [
        'date'     => ['relative', 'year', 'monthday', 'picker', 'absolute'],
        'time'     => ['absolute'],
        'datetime' => ['relative', 'countdown', 'countup']
    ];

    /**
     * @param array           $tagChildren
     * @param null|string|string[] $tagOption
     * @param array           $tag
     * @param array           $options
     * @return string|\XF\PreEscaped
     * @noinspection PhpDocMissingThrowsInspection
     */
    public function render(array $tagChildren, $tagOption, array $tag, array $options)
    {
        $content = \utf8_trim($this->renderer->renderSubTreePlain($tagChildren));

        $dateTimeObj = $this->getDateTime($content);
        if (!$dateTimeObj)
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }

        $tagOptions = $this->getTagOptions($tagOption, $options, true , ',');
        $tagOptions['content'] = $content;
        $tagOptions = $this->parseTagOptions($tagOptions, $options);

        // format options is a must
        if (!$tagOptions['format'])
        {
            return $this->renderer->renderUnparsedTag($tag, $options);
        }

        if ($this->rendererType === 'emailHtml')
        {
            return $this->templater->renderTemplate('email:sv_bbcode_time', [
                'datetime' => $dateTimeObj,
                'options'  => $tagOptions
            ]);
        }

        /** @noinspection DateTimeConstantsUsageInspection */
        $datetime = $dateTimeObj->format(\DateTime::ISO8601);
        $title = $this->getTitle($dateTimeObj, $tagOptions);

        // type option is not required at this stage
        if (\in_array($tagOptions['format'], ['countdown', 'countup'], true))
        {
             return $this->templater->renderTemplate('public:sv_bbcode_time_counter', [
                'title'            => $title,
                'nowTimestamp'     => \XF::$time,
                'otherTimestamp'   => $dateTimeObj->getTimestamp(),
                'maximumDateParts' => (int)($tagOptions['parts'] ?? 0),
                'countUp'          => $tagOptions['format'] === 'countup',
                'options'          => $tagOptions
            ]);
        }

        $text = $this->getText($dateTimeObj, $tagOptions);
        $dynamic = $this->getDynamic($dateTimeObj, $tagOptions);

        return $this->templater->renderTemplate('public:sv_bbcode_time', [
            'text'     => $text,
            'datetime' => $datetime,
            'title'    => $title,
            'dynamic'  => $dynamic,
            'options'  => $tagOptions
        ]);
    }

    /**
     * @param string $content
     * @return null|\DateTime
     * @throws \Exception
     */
    protected function getDateTime(string $content)
    {
        $timestamp = \ctype_digit($content)
            ? (int)$content
            : \strtotime($content);

        if ($timestamp === false)
        {
            return null;
        }

        $dateTime = new \DateTime();
        $dateTime->setTimestamp($timestamp);
        $dateTime->setTimezone(\XF::language()->getTimeZone());

        return $dateTime;
    }

    protected function parseUniqueTagOptions(array &$lowerTagOptions, array &$tagOptions, array &$options, array $renderingOptions, array $constraints)
    {
        $content = '';
        $type = '';
        $format = '';
        // unique options
        foreach ($lowerTagOptions as $key => $tagOption)
        {
            switch ((string)$key)
            {
                case '0';
                case 'type':
                    $type = \utf8_trim($tagOption);
                    break;
                case '1';
                case 'format':
                    $format = \utf8_trim($tagOption);
                    break;
                case '2';
                case 'parts':
                    $options['parts'] = (int)\utf8_trim($tagOption);
                    break;
                case 'content':
                    $content = $tagOption;
                    break;
            }
        }

        $options['type'] = $this->getType($content, $type);
        $options['format'] = $this->getFormat($options['type'], $format);
    }

    public function getDefaultOptions(): array
    {
        return [
            'type'   => null,
            'format' => 'relative',
            'parts' => 0
        ];
    }

    protected function getOptionConstraints(): array
    {
        // this is not a block element
        return \array_merge(parent::getOptionConstraints(), [
            'allowBlock'  => false,
            'allowFloat'  => false,
            'allowWidth'  => false,
            'allowHeight' => false,
        ]);
    }

    protected function getType(string $content, string $fallback = null): string
    {
        if (!$fallback || !isset($this->typeMap[$fallback]))
        {
            $fallback = \reset($this->typeMap);
            if (\is_array($fallback))
            {
                $fallback = \reset($fallback);
            }
        }

        if (\ctype_digit($content))
        {
            return 'datetime';
        }

        $isDate = (\strpos($content, '-') !== false);
        $isTime = (\strpos($content, ':') !== false);
        if (!$isDate && !$isTime)
        {
            return $fallback ?: 'datetime';
        }

        if (!$isDate)
        {
            return 'time';
        }

        if (!$isTime)
        {
            return 'date';
        }

        return 'datetime';
    }

    /**
     * @param string $type
     * @param string $format
     * @return string|null
     */
    protected function getFormat(string $type, string $format)
    {
        if (!\strlen($type) || !isset($this->typeMap[$type]))
        {
            $type = \reset($this->typeMap);
            if (\is_array($type))
            {
                $type = \reset($type);
            }
        }
        if (!\strlen($type))
        {
            return null;
        }

        $formats = $this->typeMap[$type] ?? [];
        if (!$formats || !\is_array($formats))
        {
            return null;
        }

        if (!\in_array($format, $formats, true))
        {
            return \reset($formats);
        }

        // special format handling
        /** @noinspection DegradedSwitchInspection */
        switch ($format)
        {
            case 'monthday':
                $format = 'monthDay';
                break;
        }

        return $format;
    }

    /**
     * @param \DateTime $dateTime
     * @param array     $tagOptions
     * @return string
     */
    protected function getText(\DateTime $dateTime, array $tagOptions): string
    {
        if ($tagOptions['format'] === 'relative')
        {
            return $this->getRelativeText($tagOptions['type'], $dateTime);
        }

        $language = \XF::language();
        switch ($tagOptions['type'])
        {
            case 'date':
                return $language->date($dateTime, $tagOptions['format']);

            case 'time':
                return $language->time($dateTime, $tagOptions['format']);
        }

        return $language->dateTime($dateTime);
    }

    protected function getRelativeText(string $type, \DateTime $dateTime): string
    {
        $language = \XF::language();

        $timestamp = $dateTime->getTimestamp();
        $timeRef = $language->getDayStartTimestamps();
        list($date, $time) = $language->getDateTimeParts($dateTime);

        if ($type === 'datetime')
        {
            if ($timestamp > ($timeRef['week'] + (7 * 86400)))
            {
                // TODO: XF bug, times are excluded from far future dates
                return $language->getDateTimeOutput($date, $time);
            }

            return $language->getRelativeDateTimeOutput($timestamp, $date, $time, true);
        }

        if ($timestamp < $timeRef['week'])
        {
            // before this week
            return $date;
        }

        if ($timestamp < $timeRef['yesterday'])
        {
            // this week
            return $language->date($timestamp, 'l');
        }

        if ($timestamp < $timeRef['today'])
        {
            // yesterday
            return $language->getPhraseText('sv_bbcode.yesterday');
        }

        if ($timestamp < $timeRef['tomorrow'])
        {
            // today
            return $language->getPhraseText('sv_bbcode.today');
        }

        if ($timestamp < ($timeRef['tomorrow'] + 86400))
        {
            // tomorrow
            return $language->getPhraseText('sv_bbcode.tomorrow');
        }

        return $date;
    }

    protected function getTitle(\DateTime $dateTime, array $tagOptions): string
    {
        $language = \XF::language();
        list($date, $time) = $language->getDateTimeParts($dateTime);

        switch ($tagOptions['type'])
        {
            case 'date':
                return $date;

            case 'time':
                return $time;
        }

        return $language->getDateTimeOutput($date, $time);
    }

    /**
     * @param \DateTime $dateTime
     * @param array     $tagOptions
     * @return array
     */
    protected function getDynamic(\DateTime $dateTime, array $tagOptions): array
    {
        if ($tagOptions['type'] !== 'datetime')
        {
            return [];
        }

        $timestamp = $dateTime->getTimestamp();
        list($date, $time) = \XF::language()->getDateTimeParts($dateTime);

        return [
            'timestamp' => $timestamp,
            'date'      => $date,
            'time'      => $time
        ];
    }
}
