<?php

namespace SV\AdvancedBbCodesPack;

/**
 * Contains event listener callbacks.
 */
class Listener
{
    /** @var string[] */
    public static $selfClosingTags = [
        'hr',
        'fab',
        'fad',
        'fal',
        'far',
        'fas',
    ];

    /**
     * @noinspection PhpDocMissingThrowsInspection
     * @param \XF\BbCode\Renderer\AbstractRenderer $renderer
     * @param string                               $type
     */
    public static function bbCodeRender(\XF\BbCode\Renderer\AbstractRenderer $renderer, string $type)
    {
        if ($renderer instanceof \XF\BbCode\Renderer\Html)
        {
            $class = \XF::app()->extendClass('SV\AdvancedBbCodesPack\BbCode\Renderer\TagRenderer');
            /** @var \SV\AdvancedBbCodesPack\BbCode\Renderer\TagRenderer $tagRenderer */
            $tagRenderer = new $class($renderer, $type);
            $tagRenderer->bindToRenderer();
        }
    }

    /**
     * @param \XF\BbCode\RuleSet $ruleSet
     * @param string|null        $context
     * @param string|null        $subContext
     * @noinspection PhpUnusedParameterInspection
     */
    public static function bbCodeRules(\XF\BbCode\RuleSet $ruleSet, string $context = null, string $subContext = null)
    {
        foreach(static::$selfClosingTags as $tag)
        {
            $ruleSet->modifyTag(\strtolower($tag), [
                'selfClosing' => true,
            ]);
        }

        $supportOptionKeys = \XF\BbCode\RuleSet::OPTION_KEYS_BOTH;
        // xtable
        $tableName = 'xtable';
        $tableParents = ['', 'td', 'th'];
        $ruleSet->modifyTag($tableName, [
            'supportOptionKeys' => $supportOptionKeys,
            'hasOption'         => null,
            'allowedParents'    => $tableParents,
            'childTags'         => [
                $tableName => [
                    'tableTag'          => true,
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => $tableParents,
                    'disableTextNodes'  => 'insideContent'
                ],
                'thead'    => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => [$tableName],
                    'disableTextNodes'  => 'inAndAfter'
                ],
                'tbody'    => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => [$tableName],
                    'disableTextNodes'  => 'inAndAfter'
                ],
                'tfoot'    => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => [$tableName],
                    'disableTextNodes'  => 'inAndAfter'
                ],
                'colgroup' => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => [$tableName],
                    'disableTextNodes'  => 'insideContent'
                ],
                'caption'  => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => [$tableName],
                ],
                'tr'       => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => [
                        $tableName,
                        'thead',
                        'tbody',
                        'tfoot'
                    ],
                    'disableTextNodes'  => 'insideContent'
                ],
                'col'      => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => ['colgroup'],
                ],
                'td'       => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => ['tr', 'th'],
                    'disableTextNodes'  => 'afterClosing'
                ],
                'th'       => [
                    'supportOptionKeys' => $supportOptionKeys,
                    'allowedParents'    => ['tr'],
                    'disableTextNodes'  => 'afterClosing'
                ]
            ]
        ]);

        $allowList = [];
        if ($ruleSet->getTag('tabs') !== null)
        {
            $allowList[] = 'tabs';
            $ruleSet->addTag('tab', [
                'allow_empty' => true,
                'validParents' => ['tabs'],
            ]);
            $ruleSet->addTag('tab_header', [
                'allow_empty' => true,
                'validParents' => ['tabs'],
            ]);
        }
        if ($ruleSet->getTag('accordion') !== null)
        {
            $allowList[] = 'accordion';
        }
        if ($ruleSet->getTag('slider') !== null)
        {
            $allowList[] = 'slider';
        }

        if ($allowList)
        {
            $ruleSet->addTag('slide', [
                'allow_empty'       => true,
                'has_option'        => 'optional',
                'supportOptionKeys' => \XF\BbCode\RuleSet::OPTION_KEYS_BOTH,
                'validParents'      => $allowList,
            ]);
            $ruleSet->addTag('slide_header', [
                'allow_empty'       => true,
                'has_option'        => 'optional',
                'supportOptionKeys' => \XF\BbCode\RuleSet::OPTION_KEYS_BOTH,
                'validParents'      => $allowList,
            ]);
        }
    }

    public static function editorTemplatePreRender(\XF\Template\Templater $templater, string &$type, string &$template, array &$params)
    {
        // legacy, left in place to prevent errors during the upgrade
    }
}