<?php

namespace SV\AdvancedBbCodesPack\Option;

use XF\App as BaseApp;
use XF\Entity\Phrase as PhraseEntity;
use XF\Finder\Phrase as PhraseFinder;
use XF\Mvc\Entity\ArrayCollection;
use XF\Mvc\Entity\Entity;
use XF\Mvc\Entity\Finder;
use XF\Mvc\Entity\Manager as EntityManager;
use XF\Option\AbstractOption;
use XF\Entity\Option as OptionEntity;
use XF\Db\AbstractAdapter as DbAdapter;

/**
 * @since 1.7.8
 */
class WordCountLabels extends AbstractOption
{
    const OPERATOR_EQUALS_TO = 'equals_to';

    const OPERATOR_NOT_EQUAL_TO = 'not_equal_to';

    const OPERATOR_GREATER_THAN = 'greater_than';

    const OPERATOR_LESSER_THAN = 'lesser_than';

    const OPERATOR_GREATER_THAN_OR_EQUAL_TO = 'greater_than_or_equal_to';

    const OPERATOR_LESSER_THAN_OR_EQUAL_TO = 'lesser_than_or_equal_to';

    public static function renderOption(
        OptionEntity $option,
        array $htmlParams
    ) : string
    {
        $choices = [];

        foreach ($option->option_value AS $choiceIndex => $choice)
        {
            $choices[] = [
                'word_count' => (int) $choice['word_count'],
                'operator' => (string) $choice['operator'],
                'label' => \XF::phrase("svAdvancedBbCode_word_count.{$choiceIndex}")
            ];
        }

        return static::getTemplate(
            'admin:svAdvancedBBCode_option_template_wordCountLabels',
            $option,
            $htmlParams,
            [
                'choices' => $choices,
                'nextCounter' => \count($choices)
            ]
        );
    }

    /**
     * @param array $value
     *
     * @return bool
     *
     * @throws \XF\PrintableException
     */
    public static function verifyOption(array &$value) : bool
    {
        $output = [];

        foreach ($value AS $label)
        {
            if (!isset($label['word_count'])
                || !isset($label['operator'])
                || !isset($label['label'])
            )
            {
                continue;
            }

            if (empty($label['label']))
            {
                continue;
            }

            $wordCount = (int)$label['word_count'];
            $map = static::buildMapValue(
                $wordCount,
                $label['operator'],
                $label['label']
            );
            if ($map && !\array_key_exists($wordCount, $output))
            {
                $output[$wordCount] = $map;
            }
        }

        \ksort($output, SORT_NUMERIC);

        $value = \array_values($output);

        static::handleLabelPhrases($value);

        return true;
    }

    protected static function isOperatorValid(string $operator) : bool
    {
        return $operator === static::OPERATOR_EQUALS_TO
            || $operator === static::OPERATOR_NOT_EQUAL_TO
            || $operator === static::OPERATOR_GREATER_THAN
            || $operator === static::OPERATOR_LESSER_THAN
            || $operator === static::OPERATOR_GREATER_THAN_OR_EQUAL_TO
            || $operator === static::OPERATOR_LESSER_THAN_OR_EQUAL_TO;
    }

    public static function buildMapValue(
        int $wordCount,
        string $operator,
        string $label
    ) : array
    {
        if (!static::isOperatorValid($operator))
        {
            return [];
        }

        return [
            'word_count' => $wordCount,
            'operator' => $operator,
            'label' => $label
        ];
    }

    /**
     * @param array $labels
     *
     * @throws \XF\PrintableException
     */
    protected static function handleLabelPhrases(array &$labels)
    {
        /** @var ArrayCollection|PhraseEntity[] $phrases */
        $phrases = static::getPhraseFinder()
            ->searchTitle('svAdvancedBbCode_word_count.', true)
            ->where('language_id', 0)
            ->where('addon_id', '') // these phrases are not associated to any add-on because
            // we do not want the phrases to be dropped when doing upgrades
            ->fetch();

        $db = static::db();
        $db->beginTransaction();

        foreach ($phrases AS $phrase)
        {
            $phrase->delete(true, false);
        }

        static::em()->clearEntityCache('XF:Phrase');

        foreach ($labels AS $labelIndex => $label)
        {
            $phrase = static::getNewPhraseEntity();
            $phrase->language_id = 0;
            $phrase->title = "svAdvancedBbCode_word_count.{$labelIndex}";
            $phrase->phrase_text = $label['label'];
            $phrase->global_cache = false;
            $phrase->addon_id = '';
            $phrase->save(true, false);

            // text is stored in phrase, not the global option
            unset($labels[$labelIndex]['label']);
        }

        $db->commit();
    }

    /**
     * @return Entity|PhraseEntity
     */
    protected static function getNewPhraseEntity() : PhraseEntity
    {
        return static::em()->create('XF:Phrase');
    }

    protected static function app() : BaseApp
    {
        return \XF::app();
    }

    protected static function em() : EntityManager
    {
        return static::app()->em();
    }

    protected static function db() : DbAdapter
    {
        return static::app()->db();
    }

    protected static function finder(string $identifier) : Finder
    {
        return static::app()->finder($identifier);
    }

    /**
     * @return Finder|PhraseFinder
     */
    protected static function getPhraseFinder() : PhraseFinder
    {
        return static::finder('XF:Phrase');
    }
}