<?php
/**
 * @noinspection PhpMissingReturnTypeInspection
 */

namespace SV\AdvancedBbCodesPack\XF\BbCode\Renderer;

use SV\AdvancedBbCodesPack\BbCode\Renderer\StyleTrait;
use function strcasecmp;

/**
 * Extends \XF\BbCode\Renderer\Html
 */
class Html extends XFCP_Html
{
    use StyleTrait;

    /** @var bool */
    protected $svRewriteTransparentAsSpoiler = false;
    /** @var bool */
    protected $svEnableTransparentColor = true;

    public function getDefaultOptions(): array
    {
        $defaultOptions = parent::getDefaultOptions();

        if ($this->getStyleProperty('svDisableLightbox'))
        {
            $defaultOptions['lightbox'] = false;
        }

        return $defaultOptions;
    }

    public function addDefaultTags()
    {
        parent::addDefaultTags();

        $options = \XF::options();
        if ($options->svTagSizeAsRem ?? false)
        {
            $this->modifyTag('size', ['callback' => 'renderTagSizeRem']);
        }

        if ($options->svTagIdentAsRem ?? false)
        {
            $this->modifyTag('ident', ['callback' => 'renderTagIndentRem']);
        }

        $this->svEnableTransparentColor = $options->svEnableTransparentColor ?? true;
        $this->svRewriteTransparentAsSpoiler = $options->svRewriteTransparentAsSpoiler ?? false;
        // XF2.2.0 to XF2.2.5 support
        if (isset($this->tags['color']['replace']))
        {
            $this->modifyTag('color', [
                'replace'  => null,
                'callback' => 'renderTagColor',
            ]);
        }
    }

    /**
     * @param array        $children
     * @param string|array $option
     * @param array        $tag
     * @param array        $options
     * @return string
     */
    public function renderTagColor(array $children, $option, array $tag, array $options)
    {
        // note; need to tag all color spans with the 'fixed-color' class so the js can check if the message text matches the background or it transparent
        if ($this->svRewriteTransparentAsSpoiler)
        {
            // do not do is-whitespace check for transparent/background color => spoiler behaviour
            $text = $this->renderSubTree($children, $options);
            $tagOption = \trim(\strtolower(\is_array($option) ? \implode(' ', $option) : (string)$option));

            if ($tagOption === 'transparent')
            {
                if (!$this->svEnableTransparentColor)
                {
                    return $text;
                }
            }
            else if (\XF::$versionId >= 2020670 && !$this->isUsableColor($tagOption))
            {
                return $text;
            }
            $color = \htmlspecialchars($tagOption);

            return $this->wrapHtml('<span class="fixed-color" style="color:'. $color .'">', $text, '</span>');
        }

        if (\XF::$versionId < 2020670)
        {
            // XF2.2.0 to XF2.2.5 support
            $text = $this->renderSubTree($children, $options);
            $tagOption = \trim(\strtolower(\is_array($option) ? \implode(' ', $option) : (string)$option));
            $color = \htmlspecialchars($tagOption);

            return $this->wrapHtml('<span style="color:' . $color . '">', $text, '</span>');
        }

        // XF2.2.6+ support
        return parent::renderTagColor($children, $option, $tag, $options);
    }

    /**
     * @param array           $children
     * @param string[]|string $option
     * @param array           $tag
     * @param array           $options
     * @return string
     * @noinspection PhpUnusedParameterInspection
     */
    public function renderTagIndentRem(array $children, $option, array $tag, array $options)
    {
        $output = $this->renderSubTree($children, $options);

        $amount = $option ? \intval($option) : 1;
        $amount = \max(1, \min($amount, 5));

        $side = \XF::language()->isRtl() ? 'right' : 'left';
        $css = 'margin-' . $side . ': ' . ($amount * 1.25) . 'rem'; // ($amount * 20) . 'px';

        $wrapped = $this->wrapHtml(
            '<p style="' . $css . '">',
            $output,
            '</p>'
        );

        return "{$wrapped}<break-start />\n";
    }

    /**
     * @param array           $children
     * @param string[]|string $option
     * @param array           $tag
     * @param array           $options
     * @return string
     * @noinspection PhpUnusedParameterInspection
     */
    public function renderTagSizeRem(array $children, $option, array $tag, array $options)
    {
        $text = $this->renderSubTree($children, $options);
        if (\trim($text) === '')
        {
            return $text;
        }

        $size = $this->getTextSizeRem($option);
        if ($size !== null)
        {
            return $this->wrapHtml(
                '<span style="font-size: ' . \htmlspecialchars($size) . '">',
                $text,
                '</span>'
            );
        }
        else
        {
            return $text;
        }
    }

    /**
     * @param string|int $inputSize
     * @return string|null
     */
    protected function getTextSizeRem($inputSize)
    {
        if (\strval(\intval($inputSize)) === \strval($inputSize))
        {
            // int only, translate size
            if ($inputSize <= 0)
            {
                return null;
            }

            switch ($inputSize)
            {
                case 1:
                    return '0.563rem';//'9px';
                case 2:
                    return '0.625rem';//'10px';
                case 3:
                    return '0.750rem';//'12px';
                case 4:
                    return '0.938rem';//'15px';
                case 5:
                    return '1.125rem';//'18px';
                case 6:
                    return '1.375rem';//'22px';
                default:
                    return '1.625rem';//'26px';
            }
        }
        else if (\preg_match('/^(\d+)(px|em|rem)$/i', $inputSize, $match))
        {
            $size = \intval($match[1]);
            $size = \max(8, \min($size, 36));
            $unit = $match[2];

            if (strcasecmp($unit, 'px') === 0)
            {
                $size = $size / 16.0;
                $unit = 'rem';
            }
            else if (strcasecmp($unit, 'em') === 0)
            {
                // em will not work as expected, so force it to be rem
                $unit = 'rem';
            }

            return $size . $unit;
        }

        return null;
    }

    /**
     * @param array           $children
     * @param string[]|string $option
     * @param array           $tag
     * @param array           $options
     * @return string
     */
    public function renderTagInlineSpoiler(array $children, $option, array $tag, array $options)
    {
        if ($options['svNestingInlineSpoiler'] ?? false)
        {
            return $this->renderSubTree($children, $options);
        }
        $options['svNestingInlineSpoiler'] = true;

        return parent::renderTagInlineSpoiler($children, $option, $tag, $options);
    }

    /**
     * @param string $content
     * @return string
     */
    protected function getRenderedInlineSpoiler($content)
    {
        if ($this->getStyleProperty('svUseNonBluredSpoiler'))
        {
            return \utf8_trim($this->templater->renderTemplate('public:sv_bbcode_spoiler', [
                'content' => $content
            ]));
        }

        return parent::getRenderedInlineSpoiler($content);
    }
}