<?php
/**
 * @noinspection PhpMissingReturnTypeInspection
 */

namespace SV\AdvancedBbCodesPack\XF\Html\Renderer;

use XF\BbCode\RuleSet;
use XF\Html\Tag;
use function array_values;

/**
 * Extends \XF\Html\Renderer\BbCode
 */
class BbCode extends XFCP_BbCode
{
    /** @var callable|null */
    protected $svOldSpanCallback;
    protected $svHeaderTags     = [];
    protected $svCssIndentRules = [
        'margin-left', // editor implements LTR indent this way
        'margin-right', // editor implements RTL indent this way
        'padding-left', // editor implements LTR indent this way
        'padding-right', // editor implements RTL indent this way
    ];

    public function __construct(array $options = [])
    {
        parent::__construct($options);
        $customBbCode = \XF::app()->bbCode()->get('custom');
        $ruleSet = \XF::app()->bbCode()->rules('base');

        if ($this->isSvExistingCustomBbCode($customBbCode, $ruleSet, 'abbr'))
        {
            $this->_handlers['abbr'] = ['filterCallback' => ['$this', 'handleSvTagAbbr']];
        }
        if ($this->isSvExistingCustomBbCode($customBbCode, $ruleSet, 'sub'))
        {
            $this->_handlers['sub'] = ['wrap' => '[SUB]%s[/SUB]'];
        }
        if ($this->isSvExistingCustomBbCode($customBbCode, $ruleSet, 'sup'))
        {
            $this->_handlers['sup'] = ['wrap' => '[SUP]%s[/SUP]'];
        }

        foreach (\range(1, 6) as $size)
        {
            $tag = 'h' . $size;
            if ($this->isSvExistingCustomBbCode($customBbCode, $ruleSet, $tag))
            {
                $this->svHeaderTags[$size] = true;
                $this->_handlers[$tag] = ['filterCallback' => ['$this', 'handleSvTagH']];
            }
        }
        if ($this->isSvExistingCustomBbCode($customBbCode, $ruleSet, 'hr'))
        {
            $this->_handlers['hr'] = ['filterCallback' => ['$this', 'handleSvTagHr'], 'skipCss' => false];
        }

        // this is special, if XF ever implements
        $this->svOldSpanCallback = $this->_handlers['span']['filterCallback'] ?? null;
        $this->_handlers['span'] = ['filterCallback' => ['$this', 'handleSvSpan'], 'skipCss' => false];
    }

    protected function isSvExistingCustomBbCode($customBbCode, RuleSet $ruleSet, string $tag): bool
    {
        return $ruleSet->getTag($tag) !== null && ($customBbCode[$tag] ?? false);
    }

    public function handleSvTagAbbr(string $text, \XF\Html\Tag $tag): string
    {
        $title = $tag->attribute('title');
        $title = $title ? '="' . $title . '"' : '';

        return '[ABBR' . $title . ']' . $text . '[/ABBR]';
    }

    /**
     * Handles heading tags.
     *
     * @param string       $text Child text of the tag
     * @param \XF\Html\Tag $tag  HTML tag triggering call
     * @return string
     */
    public function handleSvTagH(string $text, Tag $tag): string
    {
        $size = (int)\substr($tag->tagName(), 1);
        if (!($this->svHeaderTags[$size] ?? false))
        {
            return $this->handleTagH($text, $tag);
        }
        if ($size > 4)
        {
            $size = 4;
        }
        $tagText = 'H' . $size;

        return '[' . $tagText . ']' . $text . '[/' . $tagText . ']';
    }

    public function handleSvTagHr(string $text, Tag $tag): string
    {
        if ($tag->hasClass('bbc-hr'))
        {
            $class = $tag->attribute('class');
            if ($class && \preg_match('#bbc-hr--(\d+)#', $class, $matches))
            {
                $type = (int)($matches[1] ?? 0);

                return '[HR=' . $type . '][/HR]';
            }
        }

        return $this->handleTagHr($text, $tag);
    }

    protected function handleSvSpan(string $text, Tag $tag, string $preCssOutput): string
    {
        if ($tag->hasClass('bbCodeInlineSpoiler') || $tag->hasClass('bbc-spoiler'))
        {
            if ($tag->hasClass('fixed-color'))
            {
                return $text;
            }

            return '[ISPOILER]' . $text . '[/ISPOILER]';
        }

        $callable = $this->svOldSpanCallback;
        if ($callable)
        {
            return $callable($text, $tag, $preCssOutput);
        }

        return $text;
    }

    /**
     * @param \XF\Html\Tag|string|null $tag
     * @return bool
     */
    protected function isIdentTag($tag): bool
    {
        if ($tag instanceof Tag && $tag->tagName() === 'div')
        {
            $css = $tag->attribute('style');
            foreach ($this->svCssIndentRules as $cssRule)
            {
                if (!empty($css[$cssRule]))
                {
                    return true;
                }
            }
        }

        return false;
    }

    protected function svMergeChildIndentTags(Tag $tag)
    {
        /*
        https://xenforo.com/community/threads/wrapping-indent-around-quote-does-not-behave-sanely-when-toggling-bb-code-mode.200346/
        The editor is line-based, while XF's HTML is DOM based, this results in this bb-code:
            [indent]
            [code]
            1
            [/code]
            [/indent]
        When toggling bb-code mode, this will round tripping to:
            [INDENT][/INDENT]
            [INDENT][code][/INDENT]
            [INDENT]1[/INDENT]
            [INDENT][/code][/INDENT]
            [INDENT][/INDENT]
        */
        $children = $tag->children();
        $mergedTags = [];
        /** @var Tag|null $rewriteCandidate */
        $rewriteCandidate = null;
        foreach ($children as $child)
        {
            if ($rewriteCandidate == null)
            {
                if ($this->isIdentTag($child))
                {
                    $rewriteCandidate = $child;
                }
            }
            else if ($this->isIdentTag($child))
            {
                // validate the indent level is identical
                if ($rewriteCandidate->attribute('style') !== $child->attribute('style'))
                {
                    // try merging from this child node
                    $rewriteCandidate = $child;
                }
                else
                {
                    // merge into the previous tag. Preserve line-count by injecting a new line
                    $newChildTags = array_values(array_merge(
                        $rewriteCandidate->children(),
                        [new Tag('br', [], $rewriteCandidate)],
                        $child->children()
                    ));
                    \SV\AdvancedBbCodesPack\XF\Html\TagStatic::setChildren($rewriteCandidate, $newChildTags);

                    continue;
                }
            }
            else
            {
                // stop merging
                $rewriteCandidate = null;
            }

            $mergedTags[] = $child;
        }

        if (count($mergedTags) !== count($children))
        {
            \SV\AdvancedBbCodesPack\XF\Html\TagStatic::setChildren($tag, $mergedTags);
        }
    }

    public function renderChildren(Tag $tag, array $state)
    {
        $this->svMergeChildIndentTags($tag);

        return parent::renderChildren($tag, $state);
    }
}